"""
Testing the simplex sampling functionalities.
"""

import numpy as np

import pytest

from smote_variants.base import (array_array_index,
                            base_idx_neighbor_idx_simplices,
                            all_neighbor_simplices_real_idx,
                            reweight_simplex_vertices,
                            cartesian_product, vector_choice,
                            simplex_volume, simplex_volumes,
                            SimplexSamplingMixin,
                            random_samples_from_simplices,
                            add_samples)

def test_array_array_index():
    """
    Testing the array indexing.
    """
    array = np.array([[0, 1], [2, 3]])
    indices = np.array([[0, 1], [0, 1]])

    np.testing.assert_array_equal(array,
                                    array_array_index(array, indices))

def test_base_idx_neighbor_idx_simplices():
    """
    Testing the simplex generation.
    """
    all_simplices = base_idx_neighbor_idx_simplices(n_base=5,
                                                    n_neighbors=2,
                                                    n_dim=2)
    assert len(all_simplices) == 5

def test_all_neighbors_real_idx():
    """
    Testing the simplex generation in vector index coordinates.
    """
    indices = np.array([[0, 10, 11],
                        [1, 10, 11],
                        [2, 10, 11],
                        [3, 10, 11],
                        [4, 10, 11]])
    real_indices = all_neighbor_simplices_real_idx(n_dim=2,
                                                    indices=indices)

    np.testing.assert_array_equal(real_indices[0], np.array([0, 10]))
    np.testing.assert_array_equal(real_indices[1], np.array([0, 11]))

def test_simplex_volume():
    """
    Testing the volume calculation.
    """
    simplex = np.array([[0.0, 1.0], [1.0, 0.0]])
    assert simplex_volume(simplex) == np.sqrt(2.0)

def test_simplex_volumes():
    """
    Testing the volume calculation for multiple simplices.
    """
    simplices = np.array([np.array([[0.0, 1.0], [1.0, 0.0]])])
    assert simplex_volumes(simplices)[0] == np.sqrt(2.0)

    simplices3 = np.array([np.array([[0.0, 1.0], [0.0, 0.0], [1.0, 0.0]])])
    assert simplex_volumes(simplices3)[0] == 0.5

    assert len(simplex_volumes(np.array([]))) == 0

def test_reweight_simplex_vertices():
    """
    Testing the reweighting of vertices.
    """
    base_vectors = np.array([[0.0, 0.0],
                            [0.0, 1.0]])
    simplices = np.array([[0, 1]])
    vertex_weights = np.array([1.0, 0.5])

    results = reweight_simplex_vertices(base_vectors,
                                        simplices,
                                        None,
                                        vertex_weights)

    etalon = np.array([[[0.0, 0.0],
                        [0.0, 0.5]]])

    np.testing.assert_array_equal(results, etalon)

def test_cartesian_product():
    """
    Testing the Cartesian product calculation.
    """
    arr0 = np.array([0, 1])
    arr1 = np.array([3])

    cartesian = cartesian_product(arr0, arr1)

    etalon = np.array([[0, 3], [1, 3]])

    np.testing.assert_array_equal(cartesian, etalon)

def test_vector_choice():
    """
    Testing the vector choice functionality.
    """
    values = np.array([0, 1, 2, 3])
    dist = np.array([[1.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0]])

    result = vector_choice(values, dist)

    np.testing.assert_array_equal(result, np.array([0, 3]))

# note the spacial nature of this problem, the vectors [3] and [4]
# are neighbors of each other multiple times, which leads to a different
# proportion of samples when generated with the deterministic and
# random within simplex sampling
X_SSM = np.array([[0.0, 1.0], [1.0, 0.0], [0.0, 2.0],
                    [0.01, 2.00], [0.01, 2.01]])
INDICES_SSM = np.array([[0, 2, 1], [1, 0, 2], [2, 0, 1], [3, 4, 4], [4, 3, 3]])
SMALL_VOLUME_EDGE_WEIGHT_DET = 0.4
SMALL_VOLUME_EDGE_WEIGHT_RAN = 0.25
N_TO_SAMPLE_SSM = 1000

def test_simplexsamplingmixin_dim2_uniform_deterministic():
    """
    Testing uniform and deterministic sampling in 2D.
    """
    ssm = SimplexSamplingMixin(simplex_sampling='uniform',
                                within_simplex_sampling='deterministic',
                                n_dim=2,
                                random_state=5)

    assert len(ssm.get_params()) > 0

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM)

    assert len(samples) == N_TO_SAMPLE_SSM

    splitting_point = samples[samples[:, 0] == 0.0][0][1]

    # the possible splitting points for deterministic setup
    possible_splits = []
    for idx in range(2, N_TO_SAMPLE_SSM):
        step = 1.0/idx
        splits = [1.0 + step*count for count in range(1, idx)]
        possible_splits.extend(splits)

    assert splitting_point in possible_splits

    # estimating the reasonable number of samples from an edge
    frac_small_samples = np.sum(samples[:, 0] == 0.01)/N_TO_SAMPLE_SSM
    frac_diff_perc = abs(SMALL_VOLUME_EDGE_WEIGHT_DET - frac_small_samples)
    frac_diff_perc = frac_diff_perc / SMALL_VOLUME_EDGE_WEIGHT_DET

    assert frac_diff_perc < 0.05

def test_simplexsamplingmixin_dim2_uniform_random():
    """
    Testing uniform and random sampling in 2D.
    """
    ssm = SimplexSamplingMixin(simplex_sampling='uniform',
                                within_simplex_sampling='random',
                                n_dim=2,
                                random_state=5)

    assert len(ssm.get_params()) > 0

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM)

    assert len(samples) == N_TO_SAMPLE_SSM

    splitting_points = samples[samples[:, 0] == 0.0][:, 1]
    assert np.all(np.logical_and(splitting_points >= 1,
                                    splitting_points <= 2))


    # estimating the reasonable number of samples from an edge
    frac_small_samples = np.sum(samples[:, 0] == 0.01)/N_TO_SAMPLE_SSM
    frac_diff_perc = abs(SMALL_VOLUME_EDGE_WEIGHT_RAN - frac_small_samples)
    frac_diff_perc = frac_diff_perc / SMALL_VOLUME_EDGE_WEIGHT_RAN

    assert frac_diff_perc < 0.05

def test_simplexsamplingmixin_dim2_volume_deterministic():
    """
    Testing volume weighted and deterministic sampling in 2D.
    """
    ssm = SimplexSamplingMixin(simplex_sampling='volume',
                                within_simplex_sampling='deterministic',
                                n_dim=2,
                                random_state=5)

    assert len(ssm.get_params()) > 0

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM)

    assert len(samples) == N_TO_SAMPLE_SSM

    splitting_point = samples[samples[:, 0] == 0.0][0][1]

    # the possible splitting points for deterministic setup
    possible_splits = []
    for idx in range(2, N_TO_SAMPLE_SSM):
        step = 1.0/idx
        splits = [1.0 + step*count for count in range(1, idx)]
        possible_splits.extend(splits)

    assert splitting_point in possible_splits

    # estimating the reasonable number of samples from an edge
    frac_small_volume_samples = np.sum(samples[:, 0] == 0.01)/N_TO_SAMPLE_SSM

    assert frac_small_volume_samples < 0.05

def test_simplexsamplingmixin_dim2_volume_random():
    """
    Testing volume weighted and random sampling in 2D.
    """
    ssm = SimplexSamplingMixin(simplex_sampling='volume',
                                within_simplex_sampling='random',
                                n_dim=2,
                                random_state=5)

    assert len(ssm.get_params()) > 0

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM)

    assert len(samples) == N_TO_SAMPLE_SSM

    splitting_points = samples[samples[:, 0] == 0.0][:, 1]
    assert np.all(np.logical_and(splitting_points >= 1,
                                    splitting_points <= 2))


    # estimating the reasonable number of samples from an edge
    frac_small_volume_samples = np.sum(samples[:, 0] == 0.01)/N_TO_SAMPLE_SSM

    assert frac_small_volume_samples < 0.05

def test_simplexsamplingmixin_nonsense():
    """
    Testing nonsense parameters.
    """
    with pytest.raises(ValueError):
        ssm = SimplexSamplingMixin(simplex_sampling=None,
                                within_simplex_sampling='random',
                                n_dim=2,
                                random_state=5)

        ssm.sample_simplex(X=X_SSM,
                            indices=INDICES_SSM,
                            n_to_sample=N_TO_SAMPLE_SSM)

    with pytest.raises(ValueError):
        ssm = SimplexSamplingMixin(simplex_sampling='volume',
                                within_simplex_sampling=None,
                                n_dim=2,
                                random_state=5)

        ssm.sample_simplex(X=X_SSM,
                            indices=INDICES_SSM,
                            n_to_sample=N_TO_SAMPLE_SSM)

    with pytest.raises(ValueError):
        ssm = SimplexSamplingMixin(simplex_sampling='volume',
                                within_simplex_sampling='deterministic',
                                n_dim=3,
                                random_state=5)

        ssm.sample_simplex(X=np.random.random(size=(10, 2)),
                            indices=np.random.randint(5, size=(10, 5)),
                            n_to_sample=N_TO_SAMPLE_SSM)

def test_simplexsamplingmixin_dim2_base_weights():
    """
    Testing the use of base weights.
    """
    ssm = SimplexSamplingMixin(simplex_sampling='volume',
                                within_simplex_sampling='deterministic',
                                n_dim=2,
                                random_state=5)

    assert len(ssm.get_params()) > 0

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM,
                                    base_weights=np.array([0, 0, 0,
                                                        0.5, 0.5]))

    assert len(samples) == N_TO_SAMPLE_SSM
    assert np.all(samples[:, 0] == 0.01)

def test_simplexsamplingmixin_dim2_vertex_weights():
    """
    Testing the use of vertex weights.
    """
    ssm = SimplexSamplingMixin(simplex_sampling='volume',
                                within_simplex_sampling='random',
                                n_dim=2,
                                random_state=5)

    assert len(ssm.get_params()) > 0

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM,
                                    base_weights=np.array([0, 0, 0, 0.5, 0.5]),
                                    vertex_weights=np.array([1.0, 1.0, 1.0,
                                                            1.0, 0.5]))

    assert len(samples) == N_TO_SAMPLE_SSM
    samples = samples[samples[:, 0] == 0.01]
    n_above = np.sum(samples[:, 1] >= 2.005)
    n_below = np.sum(samples[:, 1] <= 2.005)

    assert abs(n_below /n_above - 3)/3 < 0.05

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM,
                                    base_weights=np.array([0, 0, 0, 1.0, 0]),
                                    vertex_weights=np.array([1.0, 1.0, 1.0,
                                                            1.0, 0.5]),
                                    X_vertices=X_SSM.copy())

    assert len(samples) == N_TO_SAMPLE_SSM
    assert np.all(np.logical_and(samples[:, 1] <= 2.005,
                                    samples[:, 1] >= 2.0))

    ssm = SimplexSamplingMixin(simplex_sampling='uniform',
                                within_simplex_sampling='random',
                                n_dim=2,
                                random_state=5)

    samples = ssm.sample_simplex(X=X_SSM,
                                    indices=INDICES_SSM,
                                    n_to_sample=N_TO_SAMPLE_SSM,
                                    vertex_weights=np.array([1.0, 1.0, 1.0,
                                                            1.0, 0.5]),
                                    X_vertices=X_SSM.copy())

    assert len(samples) == N_TO_SAMPLE_SSM

    samples = samples[samples[:, 0] == 0.01]
    n_above = np.sum(samples[:, 1] >= 2.005)
    n_below = np.sum(samples[:, 1] <= 2.005)

    assert abs(n_below /n_above - 3)/3 < 0.05

def test_random_samples_from_simplices():
    """
    Testing sampling within a triangle and checking if the samples
    fall on the correct side (or on) the hypotenuse.
    """
    X = np.array([[0, 1], [1, 0], [0, 0]])
    X_vertices = np.array([[0, 1], [1, 0], [1, 1]])
    simplices = np.array([[0, 1, 2]]*10)
    vertex_weights_0 = np.array([1.0, 1.0, 0.0])

    sample = random_samples_from_simplices(X, simplices)
    assert np.all(np.dot(sample, np.array([1.0, 1.0])) - 1 < 0)

    sample = random_samples_from_simplices(X, simplices, X_vertices=X)
    assert np.all(np.dot(sample, np.array([1.0, 1.0])) - 1 < 0)

    sample = random_samples_from_simplices(X, simplices, X_vertices=X_vertices)
    assert np.all(np.dot(sample, np.array([1.0, 1.0])) - 1 > 0)

    sample = random_samples_from_simplices(X, simplices,
                                            vertex_weights=vertex_weights_0)
    np.testing.assert_almost_equal(np.dot(sample, np.array([1.0, 1.0])) - 1, 0)

    sample = random_samples_from_simplices(X, simplices,
                                            X_vertices = X_vertices,
                                            vertex_weights=vertex_weights_0)
    np.testing.assert_almost_equal(np.dot(sample, np.array([1.0, 1.0])) - 1, 0)

def test_add_samples():
    """
    Testing the add samples functionality (deterministic sample generation)
    """
    pairs = np.array([(0, 1), (0, 2), (1, 2)])
    counts = np.array([1, 1, 2])
    count = 2
    X = np.array([[0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    X_vertices = np.array([[0.0, 2.0], [2.0, 0.0], [2.0, 2.0]])
    vertex_weights = np.array([1.0, 1.0, 0.5])

    samples = add_samples(pairs=pairs,
                            counts=counts,
                            count=count,
                            X=X,
                            X_vertices=X)

    np.testing.assert_array_equal(samples, np.array([[1.0, 1.0/3.0],
                                                    [1.0, 2.0/3.0]]))

    samples = add_samples(pairs=pairs,
                            counts=counts,
                            count=count,
                            X=X,
                            X_vertices=X_vertices)

    np.testing.assert_array_equal(samples, np.array([[1 + 1.0/3.0, 2*1.0/3.0],
                                                    [1 + 2.0/3.0, 2*2.0/3.0]]))

    samples = add_samples(pairs=pairs,
                            counts=counts,
                            count=count,
                            X=X,
                            X_vertices=X,
                            vertex_weights=vertex_weights)

    np.testing.assert_array_equal(samples, np.array([[1.0, 0.5*1.0/3.0],
                                                    [1.0, 0.5*2.0/3.0]]))

    samples = add_samples(pairs=pairs,
                            counts=counts,
                            count=count,
                            X=X,
                            X_vertices=X_vertices,
                            vertex_weights=vertex_weights)

    np.testing.assert_array_equal(samples, np.array([[1 + 0.5*1.0/3.0, 0.5*2*1.0/3.0],
                                                    [1 + 0.5*2.0/3.0, 0.5*2*2.0/3.0]]))

def test_add_gaussian():
    """
    Test the add gaussian function.
    """
    samples = np.array([[0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
    simplex_sampling = SimplexSamplingMixin()

    np.testing.assert_array_equal(samples,
                                  simplex_sampling.add_gaussian_noise(samples))



"""
Testing the G_SMOTE.
"""
import pytest

from smote_variants import G_SMOTE

from smote_variants.datasets import load_normal

def test_specific():
    """
    Oversampler specific testing
    """
    obj = G_SMOTE(method="non-linear_1.0")

    dataset = load_normal()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0

    with pytest.raises(Exception):
        G_SMOTE(method='dummy')

    with pytest.raises(Exception):
        G_SMOTE(method='non-linear_-3')



"""
Testing the AMSCO.
"""

import numpy as np

from smote_variants import AMSCO

np.random.seed(42)
X = np.random.normal(size=(20, 40))
y = np.hstack([np.repeat(1, 7), np.repeat(0, 13)])

def test_specific():
    """
    Oversampler specific testing
    """

    obj = AMSCO()
    X_samp, _ = obj.sample(X, y)

    assert len(X_samp) > 0

    particle = np.array([1, 2, 3, 4])

    for _ in range(10):
        assert len(obj.remove_elements(particle)) >= 4



"""
This module implements the AMSCO method.
"""
from dataclasses import dataclass

import numpy as np

from sklearn.model_selection import StratifiedKFold

from ..base import instantiate_obj, coalesce_dict, coalesce
from ..base import OverSampling

from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['AMSCO']

@dataclass
class AMSCOParams:
    """
    AMSCO parameters class
    """
    n_pop: int
    n_iter: int
    omega: float
    r1: float # pylint: disable=invalid-name
    r2: float # pylint: disable=invalid-name

@dataclass
class Particle:
    """
    Particle class
    """
    position: np.array
    velocity: np.array
    local_best: np.array
    local_score: tuple

@dataclass
class OSMOTEStatus:
    """
    OSMOTE status object
    """
    particles : list
    limits : list
    global_best : object
    global_score : tuple
    best_dataset : tuple

class AMSCO(OverSampling):
    """
    References:
        * BibTex::

            @article{amsco,
                        title = "Adaptive multi-objective swarm fusion for
                                    imbalanced data classification",
                        journal = "Information Fusion",
                        volume = "39",
                        pages = "1 - 24",
                        year = "2018",
                        issn = "1566-2535",
                        doi = "https://doi.org/10.1016/j.inffus.2017.03.007",
                        author = "Jinyan Li and Simon Fong and Raymond K.
                                    Wong and Victor W. Chu",
                        keywords = "Swarm fusion, Swarm intelligence
                                    algorithm, Multi-objective, Crossover
                                    rebalancing, Imbalanced data
                                    classification"
                        }

    Notes:
        * It is not clear how the kappa threshold is used, I do use the RA
            score to drive all the evolution. Particularly:

            "In the last phase of each iteration, the average Kappa value
            in current non-inferior set is compare with the latest threshold
            value, the threshold is then increase further if the average value
            increases, and vice versa. By doing so, the non-inferior region
            will be progressively reduced as the Kappa threshold lifts up."

        I don't see why would the Kappa threshold lift up if the kappa
        thresholds are decreased if the average Kappa decreases ("vice versa").

        * Due to the interpretation of kappa threshold and the lack of detailed
            description of the SIS process, the implementation is not exactly
            what is described in the paper, but something very similar.
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_memetic,
                  OverSampling.cat_uses_classifier,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 *,
                 n_pop=5,
                 n_iter=15,
                 omega=0.1,
                 r1=0.1,
                 r2=0.1,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 classifier=('sklearn.tree',
                                'DecisionTreeClassifier',
                                {'random_state': 2}),
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            n_pop (int): size of populations
            n_iter (int): optimization steps
            omega (float): intertia of PSO
            r1 (float): force towards local optimum
            r2 (float): force towards global optimum
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(random_state=random_state,
                            checks={'min_n_min': 4})
        self.check_greater_or_equal(n_pop, "n_pop", 1)
        self.check_greater_or_equal(n_iter, "n_iter", 1)
        self.check_greater_or_equal(omega, "omega", 0)
        self.check_greater_or_equal(r1, "r1", 0)
        self.check_greater_or_equal(r2, "r2", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.params = AMSCOParams(n_pop, n_iter, omega, r1, r2)
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = ss_params
        self.n_jobs = n_jobs
        self.classifier = classifier

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        # as the method is an overall optimization, 1 reasonable settings
        # should be enough

        classifiers = [('sklearn.tree',
                        'DecisionTreeClassifier',
                        {'random_state': 2})]
        parameter_combinations = {'n_pop': [5],
                                  'n_iter': [15],
                                  'omega': [0.1],
                                  'r1': [0.1],
                                  'r2': [0.1],
                                  'classifier': classifiers}

        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_scores(self, preds, tests):
        """
        Determine performance scores

        Args:
            preds (np.array): predictions
            tests (np.array): tested labels

        Returns:
            float, float: kappa, accuracy
        """
        # calculate kappa and accuracy scores
        true_p = np.sum(np.logical_and(preds == tests,
                                    tests == self.min_label))
        false_n = np.sum(np.logical_and(preds != tests,
                                    tests == self.min_label))
        true_n = np.sum(np.logical_and(preds == tests,
                                    tests == self.maj_label))
        false_p = np.sum(np.logical_and(preds != tests,
                                    tests == self.maj_label))

        p_o = (true_p + true_n)/(true_p + false_n + true_n + false_p)
        p_e = (true_p + false_n)*(true_p + false_p)/\
                (true_p + false_n + true_n + false_p)**2 + \
                (false_p + true_n)*(false_n + true_n)/\
                (true_p + false_n + true_n + false_p)**2

        kappa = (p_o - p_e)/(1.0 - p_e)
        accuracy = (true_p + true_n)/(true_p + false_n + true_n + false_p)

        return kappa, accuracy

    def fitness(self, *, X_min, X_maj, X, y, n_cross_val):
        """
        Calculating fitness function

        Args:
            X_min (np.array): minority samples
            X_maj (np.array): majority samples

        Returns:
            float, float: kappa, accuracy
        """
        kfold = StratifiedKFold(n_splits=np.max([2, np.min([n_cross_val, len(X_min), len(X_maj)])]),
                                shuffle=False)

        # prepare assembled dataset
        X_assembled = np.vstack([X_min, X_maj]) # pylint: disable=invalid-name
        y_assembled = np.hstack([np.repeat(self.min_label, len(X_min)),
                                 np.repeat(self.maj_label, len(X_maj))])

        preds = []
        tests = []

        for train, _ in kfold.split(X_assembled, y_assembled):
            classifier_obj = instantiate_obj(self.classifier)
            classifier_obj.fit(X_assembled[train], y_assembled[train])
            preds.append(classifier_obj.predict(X))
            tests.append(y)
        preds = np.hstack(preds)
        tests = np.hstack(tests)

        return self.determine_scores(preds, tests)

    def update_velocities_osmote(self, ostatus):
        """
        Update the velocities.

        Args:
            ostatus (OSMOTEStatus): status object of the search
        """
        for _, part in enumerate(ostatus.particles):
            diff1 = (part.local_best - part.velocity)
            diff2 = (ostatus.global_best - part.velocity)
            new_velocity = (part.velocity*self.params.omega +
                            self.params.r1 * diff1 +
                            self.params.r2 * diff2)
            # clipping velocities using the upper bounds of the
            # particle search space
            new_velocity[0] = np.clip(new_velocity[0],
                                        -ostatus.limits[1][0]/2,
                                        ostatus.limits[1][0]/2)
            new_velocity[1] = np.clip(new_velocity[1],
                                        -ostatus.limits[1][1]/2,
                                        ostatus.limits[1][1]/2)
            part.velocity = new_velocity

    def update_particles_osmote(self, ostatus):
        """
        Update the particles.

        Args:
            ostatus (OSMOTEStatus): status object of the search
        """
        # update particles
        for _, part in enumerate(ostatus.particles):
            new_position = part.position + part.velocity
            # clipping the particle positions using the lower and
            # upper bounds
            new_position[0] = np.clip(new_position[0],
                                        ostatus.limits[0][0],
                                        ostatus.limits[1][0])
            new_position[1] = np.clip(new_position[1],
                                        ostatus.limits[0][1],
                                        ostatus.limits[1][1])
            part.position = new_position

    def init_pop_osmote(self):
        """
        Initialize a population.

        Returns:
            np.array: the initial population
        """
        proportion = self.random_state.random_sample()/2.0+0.5
        n_neighbors = self.random_state.randint(3, 10)
        return Particle(position=np.array([proportion, n_neighbors]),
                        velocity=np.array([0.1, 1]),
                        local_best=np.array([proportion, n_neighbors]),
                        local_score=(0.0,0.0))

    def update_scores_osmote(self, ostatus, particle, score, dataset):
        """
        Update the scores based on the results for particle idx

        Args:
            ostatus (OSMOTEStatus): status object of the search
            particle (Particle): the particle
            score (tuple): score of the particle
            dataset (tuple): dataset represented by the particle
        """
        joint_local = np.prod(particle.local_score)
        joint_global = np.prod(ostatus.global_score)
        joint_score = np.prod(score)

        if joint_score > joint_local:
            particle.local_best = particle.position.copy()
            particle.local_score = score
        if joint_score > joint_global:
            ostatus.global_best = particle.position.copy()
            ostatus.global_score = score
            ostatus.best_dataset = dataset

    def evaluate_osmote(self, ostatus, *, X_maj, X_min, X, y,
                            n_cross_val, nn_params):
        """
        Evaluation

        Args:
            ostatus (OSMOTEStatus): status object of the search
            X_maj (np.array): majority samples
            X_min (np.array): minority samples
            X (np.array): all samples
            y (np.array): all targets
            n_cross_val (int): number of cross validations
            nn_params (dict): nearest neighbors parameters
        """
        # evaluate
        for _, particle in enumerate(ostatus.particles):
            # apply SMOTE
            smote = SMOTE(particle.position[0],
                            int(np.rint(particle.position[1])),
                            nn_params=nn_params,
                            ss_params=self.ss_params,
                            n_jobs=self.n_jobs,
                            random_state=self._random_state_init)

            y_to_sample = np.hstack([np.repeat(self.maj_label, len(X_maj)),
                                        np.repeat(self.min_label, len(X_min))])
            X_samp, _ = smote.sample(np.vstack([X_maj, X_min]),
                                    y_to_sample) # pylint: disable=invalid-name

            # evaluate
            score = self.fitness(X_min=X_samp[len(X_maj):],
                                    X_maj=X_samp[:len(X_maj)],
                                    X=X,
                                    y=y,
                                    n_cross_val=n_cross_val)

            # update scores according to the multiobjective setting
            self.update_scores_osmote(ostatus, particle, score,
                        dataset=(X_samp[len(X_maj):], X_samp[:len(X_maj)]))

    def osmote(self, *, X_min, X_maj, X, y, n_cross_val, nn_params):
        """
        Executing OSMOTE phase

        Args:
            X_min (np.array): minority samples
            X_maj (np.array): majority samples

        Returns:
            np.array, np.array: new minority and majority datasets
        """

        # initialize particles, first coordinate represents proportion
        # parameter of SMOTE
        # the second coordinate represents the number of neighbors to
        # take into consideration

        particles = [self.init_pop_osmote()\
                                    for _ in range(self.params.n_pop)]

        ostatus = OSMOTEStatus(
                    particles=particles,
                    limits=[np.array([0.25, 3]), np.array([4.0, 10])],
                    global_best=particles[0].position.copy(),
                    global_score=(0.0, 0.0),
                    best_dataset=None)

        # running the optimization
        for _ in range(self.params.n_iter):
            # update velocities
            self.update_velocities_osmote(ostatus)
            self.update_particles_osmote(ostatus)
            self.evaluate_osmote(ostatus=ostatus,
                                    X_maj=X_maj,
                                    X_min=X_min,
                                    X=X,
                                    y=y,
                                    n_cross_val=n_cross_val,
                                    nn_params=nn_params)

        return ostatus.best_dataset[0], ostatus.best_dataset[1]

    # initiate particles
    def init_particle_sis(self, min_num, max_num, X_maj):
        """
        Initialize particle for SIS.

        Args:
            min_num (int): minimum number
            max_num (int): maximum number
            X_maj (np.array): majority samples

        Returns:
            np.array: majority indices
        """
        num = self.random_state.randint(min_num, max_num)
        maj = self.random_state.choice(np.arange(len(X_maj)), num)
        return maj

    def remove_elements(self, particle):
        """
        Removes some random elements

        Args:
            particle (np.array): a particle

        Returns:
            np.array: the mutant particle
        """
        domain = np.arange(len(particle))
        #n_max = min([10, len(particle)])
        n_max = np.max([0, int(np.sqrt(len(particle)))])

        n_to_choose = self.random_state.randint(0, n_max)

        if len(particle) - n_to_choose < 4:
            return particle

        to_remove = self.random_state.choice(domain, n_to_choose)

        return np.delete(particle, to_remove)

    def add_elements(self, particle, mutant, X_maj):
        """
        Adds some new elements to the particle

        Args:
            particle (np.array): the particle
            mutant (np.array): the already mutated particle (
                                by removing elements)
            X_maj (np.array): the majority samples

        Returns:
            np.array: the mutated particle
        """
        maj_set = set(np.arange(len(X_maj)))
        part_set = set(particle)
        diff = list(maj_set.difference(part_set))
        n_max = min([10, len(diff)])
        n_to_choose = self.random_state.randint(0, n_max)
        diff_elements = self.random_state.choice(diff, n_to_choose)
        return np.hstack([mutant, np.array(diff_elements)])

    def sis(self, *, X_min, X_maj, X, y, n_cross_val):
        """
        SIS procedure

        Args:
            X_min (np.array): minority dataset
            X_maj (np.array): majority dataset

        Returns:
            np.array, np.array: new minority and majority datasets
        """
        min_num = len(X_min)
        max_num = len(X_maj)

        if min_num >= max_num:
            return X_min, X_maj

        particles = [self.init_particle_sis(min_num, max_num, X_maj)
                                    for _ in range(self.params.n_pop)]
        scores = [self.fitness(X_min=X_min,
                                X_maj=X_maj[particles[i]],
                                X=X,
                                y=y,
                                n_cross_val=n_cross_val)
                    for i in range(self.params.n_pop)]
        best_score = (0.0, 0.0)
        best_dataset = None

        for _ in range(self.params.n_iter):
            # mutate and evaluate
            # the way mutation or applying PSO is not described in the
            # paper in details
            for idx in range(self.params.n_pop):
                # removing some random elements
                mutant= self.remove_elements(particles[idx])
                # adding some random elements
                mutant = self.add_elements(particles[idx],
                                            mutant,
                                            X_maj)

                # evaluating the variant
                score = self.fitness(X_min=X_min,
                                    X_maj=X_maj[mutant],
                                    X=X,
                                    y=y,
                                    n_cross_val=n_cross_val)
                if score[1] > scores[idx][1]:
                    particles[idx] = mutant.copy()
                    scores[idx] = score
                if score[1] > best_score[1]:
                    best_score = score
                    best_dataset = mutant.copy()

        return X_min, X_maj[best_dataset]

    def evaluate(self, *, current, new_min, new_maj, X, y, n_cross_val):
        """
        Evaluate the actual configurations.

        Args:
            current (tuple): current best minority and majority
            new_min (np.array): new minority
            new_maj (np.array): new majority
            X (np.array): features
            y (np.array): target
            n_cross_val (int): cross validation
        """
        fitness_0 = np.prod(self.fitness(X_min=new_min,
                                            X_maj=current[1],
                                            X=X,
                                            y=y,
                                            n_cross_val=n_cross_val))
        fitness_1 = np.prod(self.fitness(X_min=current[0],
                                            X_maj=current[1],
                                            X=X,
                                            y=y,
                                            n_cross_val=n_cross_val))
        fitness_2 = np.prod(self.fitness(X_min=new_min,
                                            X_maj=new_maj,
                                            X=X,
                                            y=y,
                                            n_cross_val=n_cross_val))
        fitness_3 = np.prod(self.fitness(X_min=current[0],
                                            X_maj=new_maj,
                                            X=X,
                                            y=y,
                                            n_cross_val=n_cross_val))

        # selecting the new current_maj and current_min datasets
        _logger.info("%s: fitness scores: %f %f %f %f",
                        self.__class__.__name__, fitness_0,
                            fitness_1, fitness_2, fitness_3)

        max_fitness = np.max([fitness_0, fitness_1, fitness_2, fitness_3])
        if max_fitness in (fitness_1, fitness_3):
            current = (current[0], new_maj)
        if max_fitness in (fitness_0, fitness_2):
            current = (new_min, current[1])

        return current

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        n_cross_val = min([4, len(X_min)])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor'] = \
                        self.metric_tensor_from_nn_params(nn_params, X, y)

        # executing the main optimization procedure
        current = (X_min, X_maj)
        for iteration in range(self.params.n_iter):
            _logger.info("%s starting iteration %d",
                            self.__class__.__name__, iteration)
            new_min, _ = self.osmote(X_min=X_min,
                                        X_maj=current[1],
                                        X=X,
                                        y=y,
                                        n_cross_val=n_cross_val,
                                        nn_params=nn_params)
            _, new_maj = self.sis(X_min=current[0],
                                    X_maj=X_maj,
                                    X=X,
                                    y=y,
                                    n_cross_val=n_cross_val)

            current = self.evaluate(current=current,
                                    new_min=new_min,
                                    new_maj=new_maj,
                                    X=X,
                                    y=y,
                                    n_cross_val=n_cross_val)

        return (np.vstack([current[1], current[0]]),
                np.hstack([np.repeat(self.maj_label, len(current[1])),
                           np.repeat(self.min_label, len(current[0]))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'n_pop': self.params.n_pop,
                'n_iter': self.params.n_iter,
                'omega': self.params.omega,
                'r1': self.params.r1,
                'r2': self.params.r2,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'n_jobs': self.n_jobs,
                'classifier': self.classifier,
                **OverSampling.get_params(self)}



"""
This module implements the MCT method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict, fix_density
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['MCT']

class MCT(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{mct,
                    author = {Jiang, Liangxiao and Qiu, Chen and Li, Chaoqun},
                    year = {2015},
                    month = {03},
                    pages = {1551004},
                    title = {A Novel Minority Cloning Technique for
                                Cost-Sensitive Learning},
                    volume = {29},
                    booktitle = {International Journal of Pattern Recognition
                                    and Artificial Intelligence}
                    }

    Notes:
        * Mode is changed to median, distance is changed to Euclidean to
                support continuous features, and normalized.
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_copy]

    def __init__(self,
                 *,
                 proportion=1.0,
                 n_neighbors=5,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): additional parameter added by the developers of
                                smote-variants to support simplex sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 1, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        n_neigh = min([len(X_min), self.n_neighbors+1])
        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neigh,
                                                   n_jobs=self.n_jobs,
                                                   **nn_params)
        nearestn.fit(X_min)
        indices = nearestn.kneighbors(X_min, return_distance=False)

        # having continuous variables, the mode is replaced by median
        x_med = np.median(X_min, axis=0)
        distances = np.linalg.norm(x_med - X_min, axis=1)
        distribution = fix_density(np.sum(distances) - distances)

        # do the sampling
        samples = self.sample_simplex(X=X_min,
                                        indices=indices,
                                        n_to_sample=n_to_sample,
                                        base_weights=distribution)
        #samples = []
        #while len(samples) < n_to_sample:
        #    samples.append(X_min[self.random_state.choice(
        #        np.arange(len(X_min)), p=distribution)])

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
Testing the classification performance metrics.
"""

import numpy as np

from smote_variants.base import (prediction_labels, calculate_atoms,
                        calculate_label_scores, calculate_prob_scores,
                        calculate_all_scores)

def test_prediction_labels():
    """
    Testing the prediction label calculation.
    """
    labels = prediction_labels(np.array([0.4]))
    assert labels[0] == 0

    labels = prediction_labels(np.array([0.4, 0.6]))
    assert labels[0] == 0 and labels[1] == 1

    labels = prediction_labels(np.array([0.5, 0.5]))
    assert labels[0] == 0 and labels[1] == 1

def test_calculate_atoms():
    """
    Testing the calculation of atoms.
    """
    atoms = calculate_atoms(np.array([0, 0, 0, 0, 1, 1]),
                            np.array([0, 1, 0, 0, 1, 1]))

    assert atoms['tp'] == 2
    assert atoms['tn'] == 3
    assert atoms['fp'] == 1
    assert atoms['fn'] == 0

def test_calculate_label_scores():
    """
    Testing the calculation of label scores.
    """
    # no fn
    atoms = calculate_atoms(np.array([0, 0, 0, 0, 1, 1]),
                            np.array([0, 1, 0, 0, 1, 1]))
    atoms = calculate_label_scores(atoms)

    assert atoms['acc'] == 5/6
    assert atoms['lfn'] is None

    # no fp
    atoms = calculate_atoms(np.array([0, 1, 0, 0, 1, 1]),
                            np.array([0, 0, 0, 0, 1, 1]))
    atoms = calculate_label_scores(atoms)

    assert atoms['lfp'] is None

    # no tp
    atoms = calculate_atoms(np.array([0, 1, 0, 0, 1, 1]),
                            np.array([0, 0, 0, 0, 0, 0]))
    atoms = calculate_label_scores(atoms)

    assert atoms['ltp'] is None

    # no tn
    atoms = calculate_atoms(np.array([0, 1, 0, 0, 1, 1]),
                            np.array([1, 1, 1, 1, 1, 1]))
    atoms = calculate_label_scores(atoms)

    assert atoms['ltn'] is None

def test_calculate_prob_scores():
    """
    Testing the calculation of probability scores.
    """
    probs = np.array([0.4, 0.6])

    scores = calculate_prob_scores(np.array([0, 1]), probs)
    assert scores['auc'] == 1.0

def test_compute_all_scores():
    """
    Testing the calculation of all scores.
    """
    probs = np.array([0.4, 0.6])

    scores = calculate_all_scores(np.array([0, 1]), probs)
    assert scores['auc'] == 1.0

    probs = np.array([0.1, 0.6, 0.3, 0.9, 0.8, 0.2, 0.3, 0.9, 0.0, 1.0])
    labels = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])

    scores = calculate_all_scores(labels, probs)
    assert scores['uc'] is not None



"""
This module implements the SMOTE_Cosine method.
"""

import numpy as np

from sklearn.neighbors import NearestNeighbors

from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['SMOTE_Cosine']

class SMOTE_Cosine(OverSampling):
    """
    References:
        * BibTex::

            @article{smote_out_smote_cosine_selected_smote,
                      title={SMOTE-Out, SMOTE-Cosine, and Selected-SMOTE:
                                An enhancement strategy to handle imbalance
                                in data level},
                      author={Fajri Koto},
                      journal={2014 International Conference on Advanced
                                Computer Science and Information System},
                      year={2014},
                      pages={280-284}
                    }
    """

    categories = [OverSampling.cat_extensive]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): parameter of the NearestNeighbors component
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_formula(self, u, v, x): # pylint: disable=invalid-name
        """
        The sampling formula.

        Args:
            u (np.array): the u vector
            v (np.array): the v vector
            x (np.array): the x vector

        Returns:
            np.array: the generated samples
        """
        uu = u + self.random_state.random_sample(u.shape) * 0.3 * (u - v) # pylint: disable=invalid-name

        return x + self.random_state.random_sample(x.shape) * 0.5 * (uu - x)

    def generate_samples(self, *, X, y, X_maj, X_min,
                        n_to_sample, nn_maj_ind, composite_ind):
        """
        Generate samples

        Args:
            X (np.array): all training vectors
            y (np.array): the target labels
            X_maj (np.array): majority vectors
            X_min (np.array): minority vectors
            n_to_sample (int): number of samples to generate
            nn_maj_ind (np.array): majority neighborhood structure
            composite_ind (np.array): composite neighborhood structure

        Returns:
            np.array: the generated samples
        """
        minority_indices = np.where(y == self.min_label)[0]

        base_ind = self.random_state.choice(np.arange(len(minority_indices)),
                                            n_to_sample)

        neigh_ind = self.random_state.choice(np.arange(0, nn_maj_ind.shape[1]),
                                             n_to_sample)

        min_neigh_ind = self.random_state.choice(np.arange(1, composite_ind.shape[1]),
                                            n_to_sample)

        return self.sampling_formula(u=X[minority_indices[base_ind]],
                                     v=X_maj[nn_maj_ind[base_ind, neigh_ind]],
                                     x=X_min[composite_ind[base_ind, min_neigh_ind]])

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]
        X_maj = X[y == self.maj_label]

        # Fitting the nearest neighbors models to the minority and
        # majority data using two different metrics for the minority
        nn_min_euc = NearestNeighbors(n_neighbors=len(X_min),
                                      n_jobs=self.n_jobs)
        nn_min_euc.fit(X_min)
        nn_min_euc_ind = nn_min_euc.kneighbors(X_min, return_distance=False)

        nn_min_cos = NearestNeighbors(n_neighbors=len(X_min),
                                      metric='cosine',
                                      n_jobs=self.n_jobs)
        nn_min_cos.fit(X_min)
        nn_min_cos_ind = nn_min_cos.kneighbors(X_min, return_distance=False)

        n_neighbors = np.min([X_maj.shape[0], self.n_neighbors])
        nn_maj = NearestNeighbors(n_neighbors=n_neighbors,
                                  n_jobs=self.n_jobs)
        nn_maj.fit(X_maj)
        nn_maj_ind = nn_maj.kneighbors(X_min, return_distance=False)

        # prepare the composite neighborhood relationship
        composite_ind = (nn_min_euc_ind.argsort(axis=1) \
                            + nn_min_cos_ind.argsort(axis=1)).argsort(axis=1)

        samples = self.generate_samples(X=X, y=y, X_maj=X_maj, X_min=X_min,
                            n_to_sample=n_to_sample, nn_maj_ind=nn_maj_ind,
                            composite_ind=composite_ind)

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the TRIM_SMOTE.
"""
import numpy as np

from smote_variants import TRIM_SMOTE

from smote_variants.datasets import load_illustration_2_class

def test_specific():
    """
    Oversampler specific testing
    """

    obj = TRIM_SMOTE(min_precision = 1.0)

    dataset = load_illustration_2_class()

    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(X_samp) > 0

    assert len(obj.parameter_combinations(raw=True)) > 0

    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5],
                    [5, 6], [6, 7], [8, 9], [9, 10], [10, 11],
                    [11, 12], [12, 13], [13, 14], [14, 15],
                    [15, 16], [16, 17]])
    y = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0])

    assert obj.sample(X, y)[0].shape[0] > 0

    X = np.array([[1, 2], [1, 2], [1, 2], [1, 2], [1, 2], [1, 2],
                    [3, 4], [3, 4], [3, 4], [3, 4], [3, 4]])
    y = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

    assert obj.sample(X, y)[0].shape[0] > 0



"""
This module implements the NDO_sampling method.
"""

import numpy as np

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['NDO_sampling']

class NDO_sampling(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{ndo_sampling,
                            author={Zhang, L. and Wang, W.},
                            booktitle={2011 International Conference of
                                        Information Technology, Computer
                                        Engineering and Management Sciences},
                            title={A Re-sampling Method for Class Imbalance
                                    Learning with Credit Data},
                            year={2011},
                            volume={1},
                            number={},
                            pages={393-397},
                            keywords={data handling;sampling methods;
                                        resampling method;class imbalance
                                        learning;credit rating;imbalance
                                        problem;synthetic minority
                                        over-sampling technique;sample
                                        distribution;synthetic samples;
                                        credit data set;Training;
                                        Measurement;Support vector machines;
                                        Logistics;Testing;Noise;Classification
                                        algorithms;class imbalance;credit
                                        rating;SMOTE;sample distribution},
                            doi={10.1109/ICM.2011.34},
                            ISSN={},
                            month={Sept}}
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_ordinary,
                  OverSampling.cat_application,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 T=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            T (float): threshold parameter
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(T, "T", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = ss_params
        self.T = T # pylint: disable=invalid-name
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'T': [0.5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_alpha(self, y, ind, dist):
        """
        Determine the alpha value.

        Args:
            y (np.array): the target labels
            ind (np.array): the neighborhood structure
            dist (np.array): the distances

        Returns:
            float: the alpha value
        """
        # calculating the distances between samples in the same and different
        # classes
        indices = np.where(y[ind[:, 1:]] == self.min_label)
        if len(indices[0]) == 0:
            return 0.0
        d_intra_mean = np.mean(dist[:, 1:][indices])
        indices = np.where(y[ind[:, 1:]] == self.maj_label)
        if len(indices[0]) == 0:
            return np.inf
        d_exter_mean = np.mean(dist[:, 1:][indices])
        if d_exter_mean == 0.0:
            return np.inf

        #d_intra = []
        #d_exter = []
        #for i in range(len(X_min)):
        #    min_mask = np.where(y[ind[i][1:]] == self.min_label)[0]
        #    maj_mask = np.where(y[ind[i][1:]] == self.maj_label)[0]
        #    if len(min_mask) > 0:
        #        d_intra.append(np.mean(dist[i][1:][min_mask]))
        #    if len(maj_mask) > 0:
        #        d_exter.append(np.mean(dist[i][1:][maj_mask]))
        #d_intra_mean = np.mean(np.array(d_intra))
        #d_exter_mean = np.mean(np.array(d_exter))

        # calculating the alpha value
        alpha = d_intra_mean/d_exter_mean

        return alpha

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # fitting nearest neighbors model to find the neighbors of minority
        # samples among all elements
        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        n_neighbors = min([len(X), self.n_neighbors+1])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        dist, ind = nnmt.kneighbors(X_min)

        alpha = self.determine_alpha(y, ind, dist)

        # deciding if SMOTE is enough
        if alpha < self.T:
            smote = SMOTE(self.proportion,
                          nn_params=nn_params,
                          ss_params=self.ss_params,
                          random_state=self._random_state_init)
            return smote.sample(X, y)

        # do the sampling
        # this is a conditional sampling, which cannot be vectorized
        samples = []
        while len(samples) < n_to_sample:
            idx = self.random_state.randint(len(X_min))
            random_idx = self.random_state.choice(ind[idx][1:])
            # create sample close to the initial minority point
            samples.append(X_min[idx] + (X[random_idx] - X_min[idx])
                           * self.random_state.random_sample()/2.0)
            if y[random_idx] == self.min_label:
                # create another sample close to the neighboring minority point
                samples.append(X[random_idx] + (X_min[idx] - X[random_idx])
                               * self.random_state.random_sample()/2.0)

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'T': self.T,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the ADASYN.
"""

from smote_variants import ADASYN

from smote_variants.datasets import load_normal

def test_specific():
    """
    Oversampler specific testing
    """

    dataset = load_normal()

    obj = ADASYN(d_th=0.0)
    X_samp, _ = obj.sample(dataset['data'], dataset['target'])

    assert len(dataset['data']) == len(X_samp)



import os
import codecs

from setuptools import setup, find_packages

def readme():
    with codecs.open('README.rst', encoding='utf-8-sig') as f:
        return f.read()

version_file= os.path.join('smote_variants', '_version.py')
__version__= "0.0.0"
with open(version_file) as f:
    exec(f.read())

from setuptools.command.test import test as TestCommand
import sys

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

DISTNAME= 'smote_variants'
DESCRIPTION= 'Variants of the synthetic minority oversampling technique (SMOTE) for imbalanced learning'
LONG_DESCRIPTION= readme()
LONG_DESCRIPTION_CONTENT_TYPE='text/x-rst'
MAINTAINER= 'Gyorgy Kovacs'
MAINTAINER_EMAIL= 'gyuriofkovacs@gmail.com'
URL= 'https://github.com/analyticalmindsltd/smote_variants'
LICENSE= 'MIT'
DOWNLOAD_URL= 'https://github.com/analyticalmindsltd/smote_variants'
VERSION= __version__
CLASSIFIERS= [  'Intended Audience :: Science/Research',
                'Intended Audience :: Developers',
                'Development Status :: 3 - Alpha',
                'License :: OSI Approved :: MIT License',
                'Programming Language :: Python',
                'Topic :: Scientific/Engineering :: Artificial Intelligence',
                'Topic :: Software Development',
                'Operating System :: Microsoft :: Windows',
                'Operating System :: POSIX',
                'Operating System :: Unix',
                'Operating System :: MacOS']
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if on_rtd:
    INSTALL_REQUIRES = []
else:
    INSTALL_REQUIRES= ['numpy>=1.13.0', 'scipy', 'scikit-learn', 'joblib', 'minisom', 'statistics', 'tensorflow', 'keras', 'pandas', 'mkl', 'metric_learn', 'seaborn']
EXTRAS_REQUIRE= {'tests': ['pytest'],
                 'docs': ['sphinx', 'sphinx-gallery', 'sphinx_rtd_theme', 'matplotlib', 'pandas']}
PYTHON_REQUIRES= '>=3.5'
CMDCLASS = {'test': PyTest}
PACKAGE_DIR= {'smote_variants': 'smote_variants'}
SETUP_REQUIRES=['setuptools>=41.0.1', 'wheel>=0.33.4', 'pytest-runner']
TESTS_REQUIRE=['pytest']

setup(name=DISTNAME,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      author=MAINTAINER,
      author_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      license=LICENSE,
      url=URL,
      version=VERSION,
      download_url=DOWNLOAD_URL,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
      zip_safe=False,
      classifiers=CLASSIFIERS,
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      python_requires=PYTHON_REQUIRES,
      setup_requires=SETUP_REQUIRES,
      tests_require=TESTS_REQUIRE,
      cmdclass=CMDCLASS,
      package_dir=PACKAGE_DIR,
      packages=find_packages(exclude=[]))



"""
Testing oversampling with metric learning.
"""

import pytest
import numpy as np

import smote_variants as sv

from smote_variants.datasets import load_illustration_2_class

dataset = load_illustration_2_class()

@pytest.mark.parametrize("smote_class", sv.get_metric_learning_oversamplers())
def test_metric_learning(smote_class):
    """
    Testing oversamplers with metric learning.

    Args:
        smote_class (class): an oversampler class.
    """
    nn_params = {'metric': 'precomputed',
                 'metric_learning_method': 'ITML'}
    X, y = smote_class(nn_params=nn_params).sample(dataset['data'],
                                                   dataset['target'])

    assert np.unique(y).shape[0] == 2
    assert X.shape[0] > 0



"""
This module implements the condensed nearest neighbors technique.
"""

import numpy as np

from sklearn.neighbors import KNeighborsClassifier

from ._noisefilter import NoiseFilter

from .._logger import logger
_logger= logger

__all__= ['CondensedNearestNeighbors']


class CondensedNearestNeighbors(NoiseFilter):
    """
    Condensed nearest neighbors

    References:
        * BibTex::

            @ARTICLE{condensed_nn,
                        author={Hart, P.},
                        journal={IEEE Transactions on Information Theory},
                        title={The condensed nearest neighbor rule (Corresp.)},
                        year={1968},
                        volume={14},
                        number={3},
                        pages={515-516},
                        keywords={Pattern classification},
                        doi={10.1109/TIT.1968.1054155},
                        ISSN={0018-9448},
                        month={May}}
    """

    def __init__(self, n_jobs=1, **_kwargs):
        """
        Constructor of the noise removing object

        Args:
            n_jobs (int): number of jobs
        """
        super().__init__()

        self.check_n_jobs(n_jobs, 'n_jobs')

        self.n_jobs = n_jobs

    def get_params(self, deep=False):
        return {'n_jobs': self.n_jobs,
                **NoiseFilter.get_params(self, deep)}

    def remove_noise(self, X, y):
        """
        Removes noise from dataset

        Args:
            X (np.array): features
            y (np.array): target labels

        Returns:
            np.array, np.array: dataset after noise removal
        """
        _logger.info("%s: Running noise removal", self.__class__.__name__)

        self.class_label_statistics(y)

        # Initial result set consists of all minority samples and 1 majority
        # sample

        X_maj = X[y == self.maj_label]
        X_hat = np.vstack([X[y == self.min_label], X_maj[0]]) # pylint: disable=invalid-name
        y_hat = np.hstack([np.repeat(self.min_label, len(X_hat)-1),
                           [self.maj_label]])
        X_maj = X_maj[1:]

        # Adding misclassified majority elements repeatedly
        while len(X_maj) != 0:
            knn = KNeighborsClassifier(n_neighbors=1, n_jobs=self.n_jobs)
            knn.fit(X_hat, y_hat)
            pred = knn.predict(X_maj)

            if np.all(pred == self.maj_label):
                break

            X_hat = np.vstack([X_hat, X_maj[pred != self.maj_label]]) # pylint: disable=invalid-name
            y_hat = np.hstack([y_hat,
                                np.repeat(self.maj_label,
                                          len(X_hat) - len(y_hat))])

            X_maj = np.delete(X_maj,
                              np.where(pred != self.maj_label)[0], axis=0)

        return X_hat, y_hat



"""
This module implements the Random_SMOTE method.
"""

import numpy as np

from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['Random_SMOTE']

class Random_SMOTE(OverSampling):
    """
    References:
        * BibTex::

            @InProceedings{random_smote,
                            author="Dong, Yanjie
                            and Wang, Xuehua",
                            editor="Xiong, Hui
                            and Lee, W. B.",
                            title="A New Over-Sampling Approach: Random-SMOTE
                                    for Learning from Imbalanced Data Sets",
                            booktitle="Knowledge Science, Engineering and
                                        Management",
                            year="2011",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="343--352",
                            isbn="978-3-642-25975-3"
                            }
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_sample_componentwise,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params={},
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                    to sample e.g. 1.0 means that after
                                    sampling the number of minority samples
                                    will be equal to the number of majority
                                    samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn

        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion,
                                           self.class_stats[self.maj_label],
                                           self.class_stats[self.min_label])

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        X_min = X[y == self.min_label]

        # fitting nearest neighbors model to find closest neighbors of minority
        # points
        n_neighbors = min([len(X_min), self.n_neighbors + 1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_min)
        ind = nnmt.kneighbors(X_min, return_distance=False)

        base_indices = self.random_state.choice(np.arange(X_min.shape[0]),
                                                n_to_sample)
        y12_ind = self.random_state.choice(np.arange(1, n_neighbors),
                                            size=(n_to_sample, 2))

        y1_ind = X_min[ind[base_indices, y12_ind[:, 0]]]
        y2_ind = X_min[ind[base_indices, y12_ind[:, 1]]]

        tmp = y1_ind + (y2_ind - y1_ind) \
                        * self.random_state.random_sample(size=y1_ind.shape)

        samples = X_min[base_indices] + (tmp - X_min[base_indices]) \
                        * self.random_state.random_sample(size=tmp.shape)

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the neighborhood cleaning rule.
"""

import numpy as np

from ..base import mode, coalesce
from ._noisefilter import NoiseFilter
from ..base import (NearestNeighborsWithMetricTensor)

from .._logger import logger
_logger= logger

__all__= ['NeighborhoodCleaningRule']

class NeighborhoodCleaningRule(NoiseFilter):
    """
    References:
        * BibTex::

            @article{smoteNoise0,
                     author = {Batista, Gustavo E. A. P. A. and Prati,
                                Ronaldo C. and Monard, Maria Carolina},
                     title = {A Study of the Behavior of Several Methods for
                                Balancing Machine Learning Training Data},
                     journal = {SIGKDD Explor. Newsl.},
                     issue_date = {June 2004},
                     volume = {6},
                     number = {1},
                     month = jun,
                     year = {2004},
                     issn = {1931-0145},
                     pages = {20--29},
                     numpages = {10},
                     url = {http://doi.acm.org/10.1145/1007730.1007735},
                     doi = {10.1145/1007730.1007735},
                     acmid = {1007735},
                     publisher = {ACM},
                     address = {New York, NY, USA}
                    }
    """

    def __init__(self, nn_params=None, n_jobs=1, **_kwargs):
        """
        Constructor of the noise removal object

        Args:
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
        """
        super().__init__()

        self.check_n_jobs(n_jobs, 'n_jobs')

        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    def get_params(self, deep=False):
        return {'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **NoiseFilter.get_params(self, deep)}

    def remove_noise(self, X, y):
        """
        Removes noise

        Args:
            X (np.array): features
            y (np.array): target labels

        Returns:
            np.array, np.array: cleaned features and target labels
        """
        _logger.info("%s: Running noise removal", self.__class__.__name__)
        self.class_label_statistics(y)

        # fitting nearest neighbors with proposed parameter
        # using 4 neighbors because the first neighbor is the point itself
        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=4,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X)
        indices = nnmt.kneighbors(X, return_distance=False)

        # identifying the samples to be removed
        to_remove = []
        for idx in range(len(X)):
            if (y[idx] == self.maj_label and
                    mode(y[indices[idx][1:]]) == self.min_label):
                # if sample i is majority and the decision based on
                # neighbors is minority
                to_remove.append(idx)
            elif (y[idx] == self.min_label and
                  mode(y[indices[idx][1:]]) == self.maj_label):
                # if sample i is minority and the decision based on
                # neighbors is majority
                for jdx in indices[idx][1:]:
                    if y[jdx] == self.maj_label:
                        to_remove.append(jdx)

        # removing the noisy samples and returning the results
        to_remove = list(set(to_remove))
        return np.delete(X, to_remove, axis=0), np.delete(y, to_remove)



"""
This module implements the model selection functionalities.
"""

import os
import glob
from ast import literal_eval

from joblib import Parallel, delayed

import pandas as pd

from ..base import load_dict, instantiate_obj
from ..base import all_scores
from ._folding import Folding
from ._sampling import SamplingJob
from ._evaluation import EvaluationJob

__all__ = ['evaluate_oversamplers',
            'model_selection',
            'cross_validation',
            'execute_1_job',
            'execute_1_os',
            'execute_1_eval',
            'do_clean_up',
            'do_parse_results']

def execute_1_job(fold,
                    oversamplers,
                    scaler,
                    classifiers,
                    serialization):
    """
    Executes one oversampling and evaluation job

    Args:
        fold (dict/str): a fold
        oversamplers (list): list of oversamplers
        scaler (tuple): specification of a scaler
        classifier (list): list of classifiers
        serialization (str): serialization type ('json'/'pickle')

    Returns:
        list: the results
    """
    results = []
    for oversampler in oversamplers:
        job = SamplingJob(fold,
                            oversampler[1],
                            oversampler[2],
                            scaler=scaler,
                            cache_path=None,
                            serialization=serialization)
        result = job.do_oversampling()
        job = EvaluationJob(result,
                            classifiers,
                            cache_path=None,
                            serialization=serialization)
        results.extend(job.do_evaluation())
    return results

def execute_1_os(fold,
                    oversamplers,
                    scaler,
                    cache_path,
                    serialization):
    """
    Executes one oversampling job

    Args:
        fold (dict/str): a fold
        oversamplers (list): list of oversamplers
        scaler (tuple): specification of a scaler
        cache_path (str): the cache path
        serialization (str): serialization type ('json'/'pickle')

    Returns:
        list: the results
    """
    results = []
    for oversampler in oversamplers:
        job = SamplingJob(fold,
                            oversampler[1],
                            oversampler[2],
                            scaler=scaler,
                            cache_path=cache_path,
                            serialization=serialization)
        results.append(job.do_oversampling())
    return results

def execute_1_eval(oversampling,
                    classifiers,
                    cache_path,
                    serialization):
    """
    Executes one evaluation job

    Args:
        oversampling (dict/str): an oversampling
        classifiers (list): list of classifiers
        cache_path (str): the cache path
        serialization (str): serialization type ('json'/'pickle')

    Returns:
        list: the results
    """
    job = EvaluationJob(oversampling,
                            classifiers,
                            cache_path=cache_path,
                            serialization=serialization)
    result = job.do_evaluation()
    return result

def do_clean_up(cache_path, clean_up):
    """
    Remove unnecessary files

    Args:
        cache_path (str): the cache path
        clean_up (str): 'oversamplings'/'all'
    """

    if clean_up in ['oversamplings', 'all']:
        files = glob.glob(os.path.join(cache_path, 'oversampling*'))
        for file in files:
            os.remove(file)
    if clean_up == 'all':
        files = glob.glob(os.path.join(cache_path, 'fold*'))
        for file in files:
            os.remove(file)

def _pivot_best_scores(pdf):
    """
    Pivot the best parameters for each oversampler and classifier

    Args:
        pdf (pd.DataFrame): the results averaged over folds

    Returns:
        pd.DataFrame: the pivoted results
    """
    results_dict_classifier = {}
    results_dict_oversampler = {}
    score_means = {}
    score_stds = {}
    for score in all_scores:
        tmp = pdf[pdf[score + '_mean'] == pdf[score + '_mean'].max()]

        results_dict_classifier[f'{score}_classifier_params'] = [None]
        results_dict_classifier[f'{score}_oversampler_params'] = [None]
        score_means[f'{score}_mean'] = [None]
        score_stds[f'{score}_std'] = [None]

        if len(tmp) > 0:
            results_dict_classifier[f'{score}_classifier_params'] =\
                                     [tmp.iloc[0]['classifier_params']]
            results_dict_classifier[f'{score}_oversampler_params'] =\
                                     [tmp.iloc[0]['oversampler_params']]
            score_means[f'{score}_mean'] = [tmp.iloc[0][f'{score}_mean']]
            score_stds[f'{score}_std'] = [tmp.iloc[0][f'{score}_std']]

    return pd.DataFrame({**results_dict_classifier,
                         **results_dict_oversampler,
                         **score_means,
                         **score_stds})

def _explode_entries(pdf):
    pdf = pd.concat([pdf, pdf['scores'].apply(pd.Series)], axis=1)
    pdf['database'] = pdf['fold_descriptor'].apply(lambda x: x['name'])
    pdf['fold_idx'] = pdf['fold_descriptor'].apply(lambda x: x['fold_idx'])
    pdf['repeat_idx'] = pdf['fold_descriptor'].apply(lambda x: x['repeat_idx'])
    pdf['oversampler_params'] = pdf['oversampler']
    pdf['oversampler'] = pdf['oversampler_params'].apply(lambda x: x['class_name'])
    pdf['oversampler_params_key'] = pdf['oversampler_params'].apply(str)
    pdf['classifier_params_key'] = pdf['classifier_params'].apply(str)
    return pdf


def do_parse_results(results, parse_results, serialization):
    """
    Parse the results if needed.

    Args:
        results (list): list of results
        parse_results (bool): flag, if True, files are read
        serialization (str): serialization type 'json'/'pickle'

    Returns:
        list/pd.DataFrame: the unparsed or parsed results
    """
    if parse_results is True:
        if isinstance(results[0], str):
            loaded = []
            for res in results:
                loaded.append(load_dict(res, serialization=serialization))
            results = loaded

        pdf = pd.DataFrame(results)
        pdf_prep = _explode_entries(pdf)

        aggregations = {**{score + '_mean': (score, 'mean') for score in all_scores},
                        **{score + '_std': (score, 'std') for score in all_scores}}

        pdf_avg = pdf_prep.groupby(['database', 'oversampler',
                                'classifier', 'oversampler_params_key',
                                'classifier_module',
                                'classifier_params_key'])\
                            .agg(**aggregations)\
                            .reset_index(drop=False)
        pdf_avg['oversampler_params'] = pdf_avg['oversampler_params_key']
        pdf_avg['classifier_params'] = pdf_avg['classifier_params_key']
        pdf_avg = pdf_avg.drop(['oversampler_params_key',
                                'classifier_params_key'], axis='columns')
        return pdf_avg
    return results

def cached_evaluation(*, folding,
                        oversamplers,
                        classifiers,
                        scaler,
                        cache_path,
                        serialization,
                        n_jobs,
                        clean_up):
    """
    Executes cached evaluation

    Args:
        folding (obj): a folding object
        oversamplers (list): list of oversampling classes/objects
        classifiers (list): list of classifier objects
        scaler (tuple): scaler object
        cache_path (str): path to a cache directory
        serialization (str): serialization method 'json'/'pickle'
        n_jobs (int): number of parallel jobs
        clean_up (str): 'oversamplings'/'all'

    Returns:
        list: the results
    """
    folding.cache_foldings()

    cache_path_tmp = os.path.join(cache_path, folding.properties['name'])

    results_os = Parallel(n_jobs=n_jobs,
                        batch_size=1)(delayed(execute_1_os)(fold,
                                                    oversamplers,
                                                    scaler,
                                                    cache_path_tmp,
                                                    serialization)
                                for fold in folding.folding_files())

    results = []
    for result in results_os:
        results.extend(result)

    results = Parallel(n_jobs=n_jobs,
                        batch_size=1)(delayed(execute_1_eval)(overs,
                                                            classifiers,
                                                            cache_path_tmp,
                                                            serialization)
                                    for overs in results)

    do_clean_up(cache_path_tmp, clean_up)

    return results

def evaluate_oversamplers(datasets,
                            oversamplers,
                            classifiers,
                            *,
                            cache_path=None,
                            validator_params=None,
                            scaler=('sklearn.preprocessing',
                                    'StandardScaler',
                                    {}),
                            serialization='json',
                            clean_up='oversamplings',
                            n_jobs=1,
                            parse_results=True):
    """
    Evaluates oversampling techniques using various classifiers on
    various datasets

    Args:
        datasets (list): list of datasets and/or dataset loaders - a dataset
                            is a dict with 'data', 'target' and 'name' keys
        oversamplers (list): list of oversampling classes/objects
        classifiers (list): list of classifier objects
        cache_path (str): path to a cache directory
        validator_params (dict/None): parameters of the validator
        scaler (tuple): scaler object
        serialization (str): serialization method 'json'/'pickle'
        clean_up (str): 'oversamplings'/'all'
        n_jobs (int): number of parallel jobs
        parse_results (bool): whether to parse the results when caching happens

    Returns:
        pd.DataFrame: the evaluation results

    Example:
        import smote_variants as sv
        import imbalanced_datasets as imbd

        from sklearn.tree import DecisionTreeClassifier
        from sklearn.neighbors import KNeighborsClassifier

        datasets= [imbd.load_glass2(), imbd.load_ecoli4()]

        oversamplers= [('smote_variants', 'SMOTE_ENN', {}),
                       ('smote_variants', 'NEATER', {}),
                       ('smote_variants', 'Lee', {})]

        classifiers= [('sklearn.neighbors', 'KNeighborsClassifier',
                      {'n_neighbors': 3}),
                      ('sklearn.neighbors', 'KNeighborsClassifier',
                      {'n_neighbors': 5}),
                      ('sklearn.tree', 'DecisionTreeClassifier', {})]

        results= evaluate_oversamplers(datasets,
                                    oversamplers,
                                    classifiers)

    """

    if validator_params is None:
        validator_params = {'n_repeats': 2, 'n_splits': 5, 'random_state': 5}

    all_results = []

    for dataset in datasets:
        folding = Folding(dataset,
                            cache_path=cache_path,
                            validator_params=validator_params,
                            serialization=serialization)

        if cache_path is None:
            results = Parallel(n_jobs=n_jobs,
                                batch_size=1)(delayed(execute_1_job)(fold,
                                                                oversamplers,
                                                                scaler,
                                                                classifiers,
                                                                serialization)
                                                for fold in folding.fold())

        else:
            results = cached_evaluation(folding=folding,
                                        oversamplers=oversamplers,
                                        classifiers=classifiers,
                                        scaler=scaler,
                                        cache_path=cache_path,
                                        serialization=serialization,
                                        n_jobs=n_jobs,
                                        clean_up=clean_up)

        for result in results:
            all_results.extend(result)

    return do_parse_results(all_results, parse_results, serialization)

def model_selection(dataset,
                    oversamplers,
                    classifiers,
                    *,
                    score='auc',
                    cache_path=None,
                    validator_params=None,
                    scaler=('sklearn.preprocessing',
                            'StandardScaler',
                            {}),
                    serialization='json',
                    clean_up='oversamplings',
                    n_jobs=1):
    """
    Evaluates model selection using various classifiers on
    various datasets

    Args:
        datasets (list): list of datasets and/or dataset loaders - a dataset
                            is a dict with 'data', 'target' and 'name' keys
        oversamplers (list): list of oversampling classes/objects
        classifiers (list): list of classifier objects
        score (str): the score to drive the model selection
        cache_path (str): path to a cache directory
        validator_params (dict/None): parameters of the validator
        scaler (tuple): scaler object
        serialization (str): serialization method 'json'/'pickle'
        clean_up (str): 'oversamplings'/'all'
        n_jobs (int): number of parallel jobs
        parse_results (bool): whether to parse the results when caching happens

    Returns:
        pd.DataFrame: the evaluation results

    Example:
        import smote_variants as sv
        import imbalanced_datasets as imbd

        from sklearn.tree import DecisionTreeClassifier
        from sklearn.neighbors import KNeighborsClassifier

        datasets= [imbd.load_glass2(), imbd.load_ecoli4()]

        oversamplers= [('smote_variants', 'SMOTE_ENN', {}),
                        ('smote_variants', 'NEATER', {}),
                        ('smote_variants', 'Lee', {})]

        classifiers= [('sklearn.neighbors', 'KNeighborsClassifier',
                      {'n_neighbors': 3}),
                      ('sklearn.neighbors', 'KNeighborsClassifier',
                      {'n_neighbors': 5}),
                      ('sklearn.tree', 'DecisionTreeClassifier', {})]

        results= evaluate_oversamplers(datasets,
                                       oversamplers,
                                       classifiers)
    """
    pdf = evaluate_oversamplers(datasets=[dataset],
                                    oversamplers=oversamplers,
                                    classifiers=classifiers,
                                    cache_path=cache_path,
                                    validator_params=validator_params,
                                    scaler=scaler,
                                    serialization=serialization,
                                    clean_up=clean_up,
                                    n_jobs=n_jobs,
                                    parse_results=True)

    pdf = pdf.groupby(['database', 'oversampler',
                                    'classifier', 'classifier_module'])\
                        .apply(_pivot_best_scores)\
                        .reset_index(drop=False)\
                        .drop('level_4', axis='columns')

    best_mask = pdf[f'{score}_mean'] == pdf[f'{score}_mean'].max()
    best_row = pdf[best_mask].iloc[0]

    oversampler = ('smote_variants',
                    best_row['oversampler'],
                    literal_eval(best_row[f'{score}_oversampler_params']))
    classifier = (best_row['classifier_module'],
                    best_row['classifier'],
                    literal_eval(best_row[f'{score}_classifier_params']))

    return instantiate_obj(oversampler), instantiate_obj(classifier)

def cross_validation():
    """
    dont really know yet
    """



"""
This module implements the MOT2LD method.
"""

import warnings

import numpy as np

from sklearn.manifold import TSNE
import scipy.signal as ssignal
from scipy.linalg import circulant

from ..base import coalesce, coalesce_dict
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['MOT2LD']

class MOT2LD(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @InProceedings{mot2ld,
                            author="Xie, Zhipeng
                            and Jiang, Liyang
                            and Ye, Tengju
                            and Li, Xiaoli",
                            editor="Renz, Matthias
                            and Shahabi, Cyrus
                            and Zhou, Xiaofang
                            and Cheema, Muhammad Aamir",
                            title="A Synthetic Minority Oversampling Method
                                    Based on Local Densities in Low-Dimensional
                                    Space for Imbalanced Learning",
                            booktitle="Database Systems for Advanced
                                        Applications",
                            year="2015",
                            publisher="Springer International Publishing",
                            address="Cham",
                            pages="3--18",
                            isbn="978-3-319-18123-3"
                            }

    Notes:
        * Clusters might contain 1 elements, and all points can be filtered
            as noise.
        * Clusters might contain 0 elements as well, if all points are filtered
            as noise.
        * The entire clustering can become empty.
        * TSNE is very slow when the number of instances is over a couple
            of 1000
    """

    categories = [OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 n_components=2,
                 k=5,
                 nn_params=None,
                 ss_params=None,
                 d_cut='auto',
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_components (int): number of components for stochastic
                                neighborhood embedding
            k (int): number of neighbors in the nearest neighbor component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            d_cut (float/str): distance cut value/'auto' for automated
                                selection
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        nn_params = coalesce(nn_params, {})

        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, 'proportion', 0)
        self.check_greater_or_equal(n_components, 'n_component', 1)
        self.check_greater_or_equal(k, 'k', 1)
        if isinstance(d_cut, (int, float)):
            if d_cut <= 0:
                raise ValueError(f"{self.__class__.__name__}: Non-positive "\
                                        "d_cut is not allowed")
        elif d_cut != 'auto':
            raise ValueError(f"{self.__class__.__name__}: d_cut value "\
                                f"{d_cut} not implemented")
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_components = n_components
        self.k = k # pylint: disable=invalid-name
        self.nn_params = nn_params
        self.d_cut = d_cut
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_components': [2],
                                  'k': [3, 5, 7],
                                  'd_cut': ['auto']}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def density_peak_clustering(self, rho, indices_min, distances_min):
        """
        Do density peak clustering.

        Args:
            rho (np.array): the rho array
            indices_min (np.array): the minority neighborhood structure
            distances_min (np.array): the minority neighbor distances

        Returns:
            np.array, np.array, np.array: the r, d and idx arrays
        """
        # extracting the number of neighbors in a given radius
        closest_highest = []
        delta = []

        # implementation of the density peak clustering algorithm
        # based on http://science.sciencemag.org/content/344/6191/1492.full
        for idx, rho_val in enumerate(rho):
            closest_neighbors = indices_min[idx]
            closest_densities = rho[closest_neighbors]
            closest_highs = np.where(closest_densities > rho_val)[0]

            if len(closest_highs) > 0:
                closest_highest.append(closest_highs[0])
                delta.append(distances_min[idx][closest_highs[0]])
            else:
                closest_highest.append(-1)
                delta.append(np.max(distances_min))

        to_sort = zip(rho, delta, np.arange(len(rho)))
        r, d, idx = zip(*sorted(to_sort, key=lambda x: x[0])) # pylint: disable=invalid-name
        r, d, idx = np.array(r), np.array(d), np.array(idx) # pylint: disable=invalid-name

        return r, d, idx

    def check_enough_clusters(self,
                                d # pylint: disable=invalid-name
                                ):
        """
        Check if there are enough clusters.

        Args:
            d (np.array): clusters
        """
        if len(d) < 3:
            raise ValueError(f"Not enough clusters: {len(d)}")

    def check_enough_peaks(self, peak_indices):
        """
        Check if there are enough peaks.

        Args:
            peak_indices (np.array): the peak indices
        """
        if len(peak_indices) == 0:
            raise ValueError("No peaks found")

    def determine_cluster_centers(self, X_min, rho, indices_min, distances_min):
        """
        Determine the cluster centers.

        Args:
            X_min (np.array): the minority samples
            rho (np.array): the rho array
            indices_min (np.array): the minority neighborhood structure
            distances_min (np.array): the minority distances

        Returns:
            np.array: cluster centers
        """
        r, d, idx = self.density_peak_clustering(rho, # pylint: disable=invalid-name
                                        indices_min,
                                        distances_min)

        self.check_enough_clusters(d)

        widths = np.arange(1, int(np.rint(len(r)/2)))
        peak_indices = np.array(ssignal.find_peaks_cwt(d, widths=widths))

        self.check_enough_peaks(peak_indices)

        cluster_centers = X_min[idx[peak_indices]]

        return cluster_centers

    def determine_neighborhoods(self,
                                X_tsne, # pylint: disable=invalid-name
                                X_min,
                                nn_params):
        """
        Determine the neighborhood structures.

        Args:
            X_tsne (np.array): the transformed samples
            X_min (np.array): the minority samples
            nn_params (dict): the neighborhood parameters

        Returns:
            np.array, np.array, np.array, np.array:
                        the indices, minority distances,
                        indices and the rho array
        """
         # fitting nearest neighbors model for all training data
        n_neighbors = min([len(X_min), self.k + 1])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(X_tsne)
        distances, indices = nnmt.kneighbors(X_min)

        # fitting nearest neighbors model to the minority data
        nn_min= NearestNeighborsWithMetricTensor(n_neighbors=len(X_min),
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nn_min.fit(X_min)
        distances_min, indices_min = nn_min.kneighbors(X_min)

        if isinstance(self.d_cut, (int, float)):
            d_cut = self.d_cut
        elif self.d_cut == 'auto':
            d_cut = np.max(distances[:, 1])

        rho = nnmt.radius_neighbors(X_min, d_cut, return_distance=False)
        rho = np.array([rho_row.shape[0] for rho_row in rho])

        return indices, distances_min, indices_min, rho

    def determine_clusters(self,
                            X_tsne, # pylint: disable=invalid-name
                            y,
                            X_min,
                            nn_params):
        """
        Do the clustering.

        Args:
            X_tsne (np.array): all samples after TSNE transformation
            y (np.array): target labels
            X_min (np.array): the minority labels after transformation
            nn_params (dict): the nearest neighbors parameters

        Returns:
            np.array, np.array, np.array: cluster centers, noise mask
                                        and importance scores
        """
        indices, distances_min, indices_min, rho = \
                self.determine_neighborhoods(X_tsne, X_min, nn_params)

        cluster_centers = self.determine_cluster_centers(X_min,
                                                        rho,
                                                        indices_min,
                                                        distances_min)

        # computing local minority counts and determining noisy samples
        local_minority_count = np.sum(y[indices[:, 1:]] == self.min_label, axis=1)

        noise = np.where(np.logical_or(rho == 1, local_minority_count == 0))[0]

        return cluster_centers, noise, local_minority_count/rho

    def check_empty_clustering(self, cluster_indices):
        """
        Check if the clustering is empty

        Args:
            cluster_indices (np.array): cluster indices
        """
        # checking if clustering is empty
        empty_clustering = np.all([len(cluster) == 0 for cluster in cluster_indices])

        if empty_clustering:
            raise ValueError("Empty clustering")

    def check_probabilities(self, prob):
        """
        Check if the probabilities are empty.

        Args:
            prob (np.array): the probabilities
        """

        if np.sum(prob) == 0.0:
            raise ValueError("Empty probabilities.")

    def do_clustering(self,
                        X_tsne, # pylint: disable=invalid-name
                        y,
                        X_min):
        """
        Do the clustering.

        Args:
            X_tsne (np.array): all samples after TSNE transformation
            y (np.array): all target labels
            X_min (np.array): the minority labels after transformation

        Returns:
            np.array, np.array, np.array, np.array: the cluster indices,
                            cluster labels, probabilities and noise mask
        """
        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                self.metric_tensor_from_nn_params(nn_params, X_tsne, y)

        cluster_centers, noise, importance = \
                self.determine_clusters(X_tsne, y, X_min, nn_params)

        # finding closest cluster center to minority points and deriving
        # cluster labels
        nn_cluster= NearestNeighborsWithMetricTensor(n_neighbors=1,
                                                        n_jobs=self.n_jobs,
                                                        **(nn_params))
        nn_cluster.fit(cluster_centers)
        ind_cluster = nn_cluster.kneighbors(X_min, return_distance=False)
        cluster_labels = ind_cluster[:, 0]

        prob = importance
        prob[noise] = 0.0

        self.check_probabilities(prob)

        prob = prob/np.sum(prob)

        # extracting cluster indices
        cluster_indices = [np.where(cluster_labels == i)[0]
                           for i in range(np.max(cluster_labels) + 1)]
        # removing noise from clusters
        cluster_indices = [list(set(c).difference(set(noise)))
                           for c in cluster_indices]

        self.check_empty_clustering(cluster_indices)

        cluster_sizes = np.array([len(cluster) for cluster in cluster_indices])
        empty = np.isin(cluster_labels, np.where(cluster_sizes == 0)[0])
        prob[empty] = 0.0
        prob = prob/np.sum(prob)

        return cluster_indices, prob, noise

    def determine_cluster_probabilities(self, cluster_indices, prob):
        """
        Determine the cluster probabilities.

        Args:
            cluster_indices (np.array): the cluster indices
            prob (np.array): the sample probabilities

        Returns:
            np.array: the cluster probabilities
        """
        cluster_probs = np.zeros(len(cluster_indices))
        for idx, indices in enumerate(cluster_indices):
            if len(indices) > 0:
                cluster_probs[idx] = np.sum(prob[indices])
        cluster_probs = cluster_probs / np.sum(cluster_probs)

        return cluster_probs

    def generate_samples_in_clusters(self, *, X_min,
                            cluster_indices, prob, n_to_sample):
        """
        Generate samples in cluster.

        Args:
            X_min (np.array): the minority samples
            cluster_indices (np.array): the cluster indices
            prob (np.array): the probabilities
            n_to_sample (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        cluster_probs = self.determine_cluster_probabilities(cluster_indices,
                                                                prob)

        clusters_selected = self.random_state.choice(len(cluster_probs),
                                                    n_to_sample,
                                                    p=cluster_probs)
        cluster_unique, cluster_count = np.unique(clusters_selected,
                                                    return_counts=True)

        #n_dim_original = self.n_dim
        samples = []
        for idx, cluster in enumerate(cluster_unique):
            cluster_vectors = X_min[cluster_indices[cluster]]
            within_prob = prob[cluster_indices[cluster]]
            within_prob = within_prob / np.sum(within_prob)
            #self.n_dim = np.min([self.n_dim, cluster_vectors.shape[0]])
            samples.append(self.sample_simplex(X=cluster_vectors,
                            indices=circulant(np.arange(cluster_vectors.shape[0])),
                            n_to_sample=cluster_count[idx],
                            base_weights=within_prob))
            #self.n_dim = n_dim_original

        #samples = []
        #while len(samples) < n_to_sample:
        #    # random sample according to the distribution computed
        #    random_idx = self.random_state.choice(np.arange(len(X_min)),
        #                                          p=prob)
        #
        #    # cluster label of the random minority sample
        #    cluster_label = cluster_labels[random_idx]
        #    if cluster_label == -1:
        #        continue
        #
        #    if len(cluster_indices[cluster_label]) == 0:
        #        continue
        #    elif len(cluster_indices[cluster_label]) == 1:
        #        # if the cluster has only 1 elements, it is repeated
        #        samples.append(X_min[random_idx])
        #        continue
        #
        #    # otherwise a random cluster index is selected for sample
        #    # generation
        #    clus = cluster_indices[cluster_label]
        #    random_neigh_in_clus_idx = self.random_state.choice(clus)
        #    while random_idx == random_neigh_in_clus_idx:
        #        random_neigh_in_clus_idx = self.random_state.choice(clus)
        #
        #    X_rand = X_min[random_idx]
        #    X_in_clus = X_min[random_neigh_in_clus_idx]
        #    samples.append(self.sample_between_points(X_rand, X_in_clus))

        return np.vstack(samples)

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        _logger.info("%s: Starting TSNE n: %d d: %d",
                        self.__class__.__name__, len(X), len(X[0]))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            # do the stochastic embedding
            X_tsne = TSNE(self.n_components, # pylint: disable=invalid-name
                        random_state=self.random_state,
                        perplexity=np.min([10, X.shape[0]-1]),
                        n_iter_without_progress=100,
                        n_iter=500,
                        verbose=0).fit_transform(X)

        X_min = X_tsne[y == self.min_label]
        _logger.info("%s: TSNE finished", self.__class__.__name__)

        try:
            cluster_indices, prob, noise = \
                self.do_clustering(X_tsne, y, X_min)
        except ValueError as valueerror:
            return self.return_copies(X, y, valueerror.args[0])

        # carrying out the sampling
        X_min = X[y == self.min_label]

        samples = self.generate_samples_in_clusters(X_min=X_min,
                                                cluster_indices=cluster_indices,
                                                prob=prob,
                                                n_to_sample=n_to_sample)

        return (np.vstack([np.delete(X, noise, axis=0), np.vstack(samples)]),
                np.hstack([np.delete(y, noise),
                           np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_components': self.n_components,
                'k': self.k,
                'nn_params': self.nn_params,
                'd_cut': self.d_cut,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the ANS method.
"""

import numpy as np

from ..base import coalesce_dict, coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['ANS']

class ANS(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @article{ans,
                     author = {Siriseriwan, W and Sinapiromsaran, Krung},
                     year = {2017},
                     month = {09},
                     pages = {565-576},
                     title = {Adaptive neighbor synthetic minority OverSamplingSimplexSimplex
                                technique under 1NN outcast handling},
                     volume = {39},
                     booktitle = {Songklanakarin Journal of Science and
                                    Technology}
                     }

    Notes:
        * The method is not prepared for the case when there is no c satisfying
            the condition in line 25 of the algorithm, fixed.
        * The method is not prepared for empty Pused sets, fixed.
    """
    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_density_based,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 nn_params=None,
                 ss_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                    to sample e.g. 1.0 means that after
                                    sampling the number of minority samples
                                    will be equal to the number of majority
                                    samples
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling params
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.
        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [
            0.1, 0.25, 0.5, 0.75, 1.0, 1.5, 2.0]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def determine_C(self, out_border, X): # pylint: disable=invalid-name
        """
        Determines the best C value.

        Args:
            out_border (np.array): outer border
            X (np.array): feature vectors

        Returns:
            float: the best C value
        """
        C_max = int(0.25*len(X)) # pylint: disable=invalid-name
        n_oc_m1 = -1
        C = 0 # pylint: disable=invalid-name
        best_diff = np.inf
        for c in range(1, C_max): # pylint: disable=invalid-name
            n_oc = np.sum(out_border >= c)
            if abs(n_oc - n_oc_m1) < best_diff:
                best_diff = abs(n_oc - n_oc_m1)
                C = n_oc # pylint: disable=invalid-name
            n_oc_m1 = n_oc

        return C

    def determine_first_pos_neighbor_distances(self, X_min, nn_params):
        """
        Determine the first positive nearest neighbor distances.

        Args:
            np.array (X_min): minority (positive) samples
            nn_params (dict): nearest neighbor parameters

        Returns:
            np.array: the first positive neighbor distances
        """
        # finding the first minority neighbor of minority samples
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=2,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nearestn.fit(X_min)
        dist, _ = nearestn.kneighbors(X_min)

        # extracting the distances of first minority neighbors from minority
        # samples
        first_pos_neighbor_distances = dist[:, 1]

        return first_pos_neighbor_distances

    def determine_outer_border(self, *, X, y, X_min,
                                first_pos_neighbor_distances,
                                nn_params):
        """
        Determine the outer border

        Args:
            X (np.array): features
            y (np.array): targets
            X_min (np.array): minority samples
            first_pos_neighbor_distances (np.array): closest positive neighbor
                                                        distances
            nn_params (dict): nearest neighbors parameters

        Returns:
            np.array: the outer border
        """
        # fitting another nearest neighbors model to extract majority
        # samples in the neighborhoods of minority samples
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=1,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nearestn.fit(X)

        # extracting the number of majority samples in the neighborhood of
        # minority samples
        out_border = []
        for idx, row in enumerate(X_min):
            ind = nearestn.radius_neighbors(row.reshape(1, -1),
                                      first_pos_neighbor_distances[idx],
                                      return_distance=False)
            out_border.append(np.sum(y[ind[0].astype(int)] == self.maj_label))

        out_border = np.array(out_border)

        return out_border

    def determine_Pused(self, X, y, X_min, nn_params): # pylint: disable=invalid-name
        """
        Determine the Pused array.

        Args:
            X (np.array): all features
            y (np.array): all targets
            X_min (np.array): minority samples
            nn_params (dict): nearest neighbor parameters

        Returns:
            np.array, float: the Pused array, and the eps value
        """
        first_pos_dist = \
            self.determine_first_pos_neighbor_distances(X_min, nn_params)

        out_border = self.determine_outer_border(X=X, y=y, X_min=X_min,
                                first_pos_neighbor_distances=first_pos_dist,
                                nn_params=nn_params)

        # finding the optimal C value by comparing the number of outcast
        # minority samples when traversing the range [1, C_max]
        # maximum C value
        C = self.determine_C(out_border, X) # pylint: disable=invalid-name

        # determining the set of minority samples Pused
        Pused = np.where(out_border < C)[0] # pylint: disable=invalid-name

        # finding the maximum distances of first positive neighbors
        if Pused.shape[0] == 0:
            eps = None
        else:
            eps = np.max(first_pos_dist[Pused])

        return Pused, eps

    def determine_simplex_weights(self,
                                    ind,
                                    nearestn,
                                    X_min,
                                    Pused # pylint: disable=invalid-name
                                    ):
        """
        Determine simplex weights.

        Args:
            ind (np.array): neighborhood structure
            nearestn (NearestNeighbors): a fitted nearest neighbors object
            X_min (np.array): minority samples
            Pused (np.array): the Pused array

        Returns:
            weights (np.array): the simplex node weights for sampling
        """
        max_neighbors = np.max([len(row) for i, row in enumerate(ind)])
        ind_dense = nearestn.kneighbors(X_min[Pused], max_neighbors, return_distance=False)

        weights = ind_dense.copy()
        weights[:,:]= 1.0

        for idx, row in enumerate(ind):
            weights[idx, len(row):] = 0.0

        return ind_dense, weights

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed.")

        X_min = X[y == self.min_label]

        # outcast extraction algorithm

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(
                                                            nn_params, X, y)

        Pused, eps = self.determine_Pused(X, y, X_min, nn_params) # pylint: disable=invalid-name

        # Adaptive neighbor SMOTE algorithm

        # checking if there are minority samples left
        if len(Pused) == 0:
            return self.return_copies(X, y, "Pused is empty")

        # fitting nearest neighbors model to find nearest minority samples in
        # the neighborhoods of minority samples
        nearestn = NearestNeighborsWithMetricTensor(n_neighbors=1,
                                                    n_jobs=self.n_jobs,
                                                    **(nn_params))
        nearestn.fit(X_min[Pused])
        ind = nearestn.radius_neighbors(X_min[Pused],
                                        eps,
                                        return_distance=False)

        # extracting the number of positive samples in the neighborhoods
        Np = np.array([len(i) for i in ind]) # pylint: disable=invalid-name

        if np.all(Np == 1):
            return self.return_copies(X, y, "all samples have only 1 neighbor"\
                                                        " in the given radius")

        # determining the distribution used to generate samples
        Np = Np/np.sum(Np) # pylint: disable=invalid-name

        ind_dense, weights = self.determine_simplex_weights(ind, nearestn,
                                                            X_min, Pused)

        samples = self.sample_simplex(X=X_min[Pused],
                                        indices=ind_dense,
                                        n_to_sample=n_to_sample,
                                        simplex_weights=weights,
                                        base_weights=Np)

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, samples.shape[0])]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}



"""
This module implements the DSRBF method.
"""
import warnings

import numpy as np

from sklearn.preprocessing import StandardScaler
from sklearn.cluster import KMeans

from ..config import suppress_external_warnings
from ..base import OverSampling
from ..base import RandomStateMixin, coalesce, coalesce_dict
from ._smote import SMOTE

from .._logger import logger
_logger= logger

__all__= ['DSRBF']

class RBFNeuron(RandomStateMixin):
    """
    This class abstracts a neuron of an RBF network
    """

    def __init__(self,
                 *,
                 center,
                 i_b,
                 o_b,
                 ranges,
                 range_mins,
                 init_conn_mask,
                 init_conn_weights,
                 beta=None,
                 mask=None,
                 input_weights=None,
                 radius=None,
                 random_state=None):
        """
        Constructor of the neuron

        Args:
            center (np.array): center of the hidden unit
            i_b (float): upper bound on the absolute values of input weights
            o_b (float): upper bound on the absolute values of output weights
            ranges (np.array): ranges widths of parameters
            range_min (np.array): lower bounds of parameter ranges
            init_conn_mask (np.array): initial input connections
            init_conn_weights (np.array): initial weights of input connections
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        RandomStateMixin.__init__(self, random_state)
        self.center = center

        self.beta = coalesce(beta, (self.random_state.random_sample()-0.5) * o_b)
        self.init_conn_mask = coalesce(mask, init_conn_mask)
        self.input_weights = coalesce(input_weights, init_conn_weights)
        self.mask = init_conn_mask
        self.radius = coalesce(radius, self.random_state.random_sample())

        self.params = {'i_b': i_b,
                        'o_b': o_b,
                        'init_conn_mask': init_conn_mask,
                        'init_conn_weights': init_conn_weights,
                        'ranges': ranges,
                        'range_mins': range_mins,
                        'beta': beta,
                        'input_weights': input_weights
                        }

    def clone(self):
        """
        Clones the neuron

        Returns:
            RBFNeuron: an identical neuron
        """
        return RBFNeuron(**(self.params),
                        center=self.center,
                        random_state=self._random_state_init,
                        mask=self.mask.copy(),
                        radius=self.radius)

    def evaluate(self, X):
        """
        Evaluates the system on dataset X

        Args:
            X (np.array): dataset to evaluate on

        Returns:
            np.array: the output of the network
        """
        weighted = X[:, self.mask]*self.input_weights
        shifted = weighted - self.center[self.mask]
        term_exp = -np.linalg.norm(shifted, axis=1)**2/self.radius**2
        return self.beta*np.exp(term_exp)

    def mutate(self):
        """
        Mutates the neuron
        """
        n_dim = self.center.shape[0]
        rand = self.random_state.random_sample()
        if rand < 0.2:
            # centre creep
            self.center = self.random_state.normal(self.center,
                                                    self.radius)
        elif rand < 0.4:
            # radius creep
            tmp = self.random_state.normal(self.radius,
                                            np.var(self.params['ranges']))
            if tmp > 0:
                self.radius = tmp
        elif rand < 0.6:
            # randomize centers
            self.center = self.random_state.random_sample(size=n_dim)\
                            * self.params['ranges'] + self.params['range_mins']
        elif rand < 0.8:
            # randomize radii
            self.radius = self.random_state.random_sample()\
                             * np.mean(self.params['ranges'])
        else:
            # randomize output weight
            self.beta = self.random_state.normal(self.beta, self.params['o_b'])

    def add_connection(self):
        """
        Adds a random input connection to the neuron
        """
        n_dim = self.center.shape[0]
        if len(self.mask) < n_dim:
            d_set = set(range(n_dim))
            mask_set = set(self.mask.tolist())

            domain = list(d_set.difference(mask_set))

            additional_elements = np.array(self.random_state.choice(domain))

            self.mask = np.hstack([self.mask,
                                    additional_elements])

            random_weight = (self.random_state.random_sample() - 0.5)\
                                                     * self.params['i_b']

            self.input_weights = np.hstack([self.input_weights,
                                            random_weight])

    def delete_connection(self):
        """
        Deletes a random input connection
        """
        if len(self.mask) > 1:
            idx = self.random_state.randint(len(self.mask))
            self.mask = np.delete(self.mask, idx)
            self.input_weights = np.delete(self.input_weights, idx)

class RBF(RandomStateMixin):
    """
    RBF network abstraction
    """

    def __init__(self,
                 *,
                 X,
                 m_min,
                 m_max,
                 i_b,
                 o_b,
                 init_conn_mask,
                 init_conn_weights,
                 range_mins=None,
                 ranges=None,
                 neurons=None,
                 beta_0=None,
                 random_state=None):
        """
        Initializes the RBF network

        Args:
            X (np.array): dataset to work with
            m_min (int): minimum number of hidden neurons
            m_max (int): maximum number of hidden neurons
            Ib (float): maximum absolute value of input weights
            Ob (float): maximum absolute value of output weights
            init_conn_mask (np.array): initial input connections
            init_conn_weights (np.array): initial input weights
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        RandomStateMixin.__init__(self, random_state)

        self.X = X

        self.range_mins = coalesce(range_mins, np.min(X, axis=0))
        self.ranges = coalesce(ranges, np.max(X, axis=0) - self.range_mins)

        beta_0 = coalesce(beta_0, (self.random_state.random_sample() - 0.5) * o_b)

        self.params = {'m_min': m_min,
                        'm_max': m_max,
                        'i_b': i_b,
                        'o_b': o_b,
                        'init_conn_mask': init_conn_mask,
                        'init_conn_weights': init_conn_weights,
                        'beta_0': beta_0}

        if neurons is not None:
            self.neurons = neurons
        else:
            self.neurons = []
            num_neurons = self.random_state.randint(m_min, m_max)
            for _ in range(num_neurons):
                self.neurons.append(self.create_new_node())

    def clone(self):
        """
        Clones the entire network

        Returns:
            RBF: the cloned network
        """
        return RBF(X=self.X,
                    **self.params,
                    neurons=[n.clone() for n in self.neurons],
                    range_mins=self.range_mins.copy(),
                    ranges=self.ranges.copy(),
                    random_state=self._random_state_init)

    def create_new_node(self):
        """
        Creates a new node.

        Returns:
            RBFNeuron: a new hidden neuron
        """
        return RBFNeuron(center=self.X[self.random_state.randint(self.X.shape[0])],
                         i_b=self.params['i_b'],
                         o_b=self.params['o_b'],
                         ranges=self.ranges,
                         range_mins=self.range_mins,
                         init_conn_mask=self.params['init_conn_mask'],
                         init_conn_weights=self.params['init_conn_weights'],
                         random_state=self._random_state_init)

    def update_data(self, X):
        """
        Updates the data to work with

        Returns:
            RBF: the updated RBF object
        """
        self.X = X
        for neuron in self.neurons:
            neuron.X = X
        return self

    def improve_centers(self):
        """
        Improves the center locations by kmeans clustering

        Returns:
            RBF: the RBF object with updated centers
        """
        if len(np.unique(self.X, axis=0)) > len(self.neurons):
            cluster_init = np.vstack([neuron.center for neuron in self.neurons])
            kmeans = KMeans(n_clusters=len(self.neurons),
                            init=cluster_init,
                            n_init=1,
                            max_iter=30,
                            random_state=self._random_state_init)
            with warnings.catch_warnings():
                if suppress_external_warnings():
                    warnings.simplefilter("ignore")
                kmeans.fit(self.X)

            for idx, neuron in enumerate(self.neurons):
                neuron.center = kmeans.cluster_centers_[idx]

        return self

    def evaluate(self, X, y):
        """
        Evaluates the target function

        Returns:
            float: the target function value
        """
        evaluation = np.column_stack([n.evaluate(X) for n in self.neurons])
        func = self.params['beta_0'] + np.sum(evaluation, axis=1)
        L_star = np.mean(abs(y[y == 1] - func[y == 1])) # pylint: disable=invalid-name
        L_star += np.mean(abs(y[y == 0] - func[y == 0])) # pylint: disable=invalid-name
        return L_star

    def mutation(self):
        """
        Mutates the neurons

        Returns:
            RBF: a new, mutated RBF network
        """
        rbf = self.clone()
        for neuron in rbf.neurons:
            neuron.mutate()
        return rbf

    def structural_mutation(self):
        """
        Applies structural mutation

        Returns:
            RBF: a new, structurally mutated network
        """
        # in the binary case the removal of output connections is the same as
        # removing hidden nodes
        rbf = self.clone()
        rand = self.random_state.random_sample()
        if rand < 0.5:
            if len(rbf.neurons) < rbf.params['m_max']:
                rbf.neurons.append(rbf.create_new_node())
            elif len(rbf.neurons) > rbf.params['m_min']:
                del rbf.neurons[self.random_state.randint(len(rbf.neurons))]
        else:
            rbf.neurons[self.random_state.randint(len(rbf.neurons))]\
                                                        .delete_connection()
            rbf.neurons[self.random_state.randint(len(rbf.neurons))]\
                                                        .add_connection()

        return rbf

    def recombine(self, rbf):
        """
        Recombines two networks

        Args:
            rbf (RBF): another network

        Returns:
            RBF: the result of recombination
        """
        # the order of neurons doesn't matter, so the logic can be simplified
        new_rbf = self.clone()

        if self.random_state.random_sample() < 0.5:
            n_random = self.random_state.randint(1, len(new_rbf.neurons))
            new_neurons_0 = self.random_state.choice(new_rbf.neurons, n_random)

            n_random = self.random_state.randint(1, len(rbf.neurons))
            new_neurons_1 = self.random_state.choice(rbf.neurons, n_random)

            new_rbf.neurons = [n.clone() for n in new_neurons_0]
            new_rbf.neurons.extend([n.clone() for n in new_neurons_1])

            while len(new_rbf.neurons) > self.params['m_max']:
                n_neurons = len(new_rbf.neurons)
                del new_rbf.neurons[self.random_state.randint(n_neurons)]
        else:
            for idx, _ in enumerate(new_rbf.neurons):
                if self.random_state.random_sample() < 0.2:
                    n_random = self.random_state.randint(len(rbf.neurons))
                    new_rbf.neurons[idx] = rbf.neurons[n_random].clone()

        return new_rbf


class DSRBF(OverSampling):
    """
    References:
        * BibTex::

            @article{dsrbf,
                        title = "A dynamic over-sampling procedure based on
                                    sensitivity for multi-class problems",
                        journal = "Pattern Recognition",
                        volume = "44",
                        number = "8",
                        pages = "1821 - 1833",
                        year = "2011",
                        issn = "0031-3203",
                        doi = "https://doi.org/10.1016/j.patcog.2011.02.019",
                        author = "Francisco Fernández-Navarro and César
                                    Hervás-Martínez and Pedro Antonio
                                    Gutiérrez",
                        keywords = "Classification, Multi-class, Sensitivity,
                                    Accuracy, Memetic algorithm, Imbalanced
                                    datasets, Over-sampling method, SMOTE"
                        }

    Notes:
        * It is not entirely clear why J-1 output is supposed where J is the
            number of classes.
        * The fitness function is changed to a balanced mean loss, as I found
            that it just ignores classification on minority samples
            (class label +1) in the binary case.
        * The iRprop+ optimization is not implemented.
        * The original paper proposes using SMOTE incrementally. Instead of
            that, this implementation applies SMOTE to generate all samples
            needed in the sampling epochs and the evolution of RBF networks
            is used to select the sampling providing the best results.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_uses_classifier,
                  OverSampling.cat_sample_ordinary,
                  OverSampling.cat_memetic,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 ss_params=None,
                 hidden_range=(4, 10),
                 i_b=2,
                 o_b=2,
                 n_pop=500,
                 n_init_pop=5000,
                 n_iter=40,
                 n_sampling_epoch=5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal
                                to the number of majority samples
            n_neighbors (int): number of neighbors in the SMOTE sampling
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            ss_params (dict): simplex sampling parameters
            m_min (int): minimum number of hidden units
            m_max (int): maximum number of hidden units
            i_b (float): input weight range
            o_b (float): output weight range
            n_pop (int): size of population
            n_init_pop (int): size of initial population
            n_iter (int): number of iterations
            n_sampling_epoch (int): resampling after this many iterations
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)

        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(hidden_range[0], "m_min", 1)
        self.check_greater_or_equal(hidden_range[1], "m_max", 1)
        self.check_greater(i_b, "Ib", 0)
        self.check_greater(o_b, "Ob", 0)
        self.check_greater_or_equal(n_pop, "n_pop", 2)
        self.check_greater_or_equal(n_init_pop, "n_pop", 2)
        self.check_greater_or_equal(n_iter, "n_iter", 0)
        self.check_greater_or_equal(n_sampling_epoch, "n_sampling_epoch", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.ss_params = coalesce_dict(ss_params, {'n_dim': 2,
                                        'simplex_sampling': 'uniform',
                                        'within_simplex_sampling': 'random',
                                        'gaussian_component': None})
        self.params = {'m_min': hidden_range[0],
                        'm_max': hidden_range[1],
                        'i_b': i_b,
                        'o_b': o_b}
        self.search_params = {'n_pop': n_pop,
                                'n_init_pop': n_init_pop,
                                'n_iter': n_iter,
                                'n_sampling_epoch': n_sampling_epoch}
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        # as the technique optimizes, it is unnecessary to check various
        # combinations except one specifying a decent workspace with a large
        # number of iterations
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7],
                                  'hidden_range': [(4, 10)],
                                  'i_b': [2.0],
                                  'o_b': [2.0],
                                  'n_pop': [100],
                                  'n_init_pop': [1000],
                                  'n_iter': [40],
                                  'n_sampling_epoch': [8]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def init_pop(self, X, domain, n_random, m_max):
        """
        Initialize an element of the population.

        Args:
            X (np.array): input data
            domain (np.array): domain of connections
            n_random (np.array): the random cardinality of connections
            m_max (int): maximum number of hidden units
        """
        init_conn_mask = self.random_state.choice(domain, n_random)
        init_conn_weights = self.random_state.random_sample(size=n_random)

        return RBF(X=X,
                    m_min=self.params['m_min'],
                    m_max=m_max,
                    i_b=self.params['i_b'],
                    o_b=self.params['o_b'],
                    init_conn_mask=init_conn_mask,
                    init_conn_weights=init_conn_weights,
                    random_state=self._random_state_init)

    def recombination(self, population):
        """
        Do the recombination.

        Args:
            population (list): the population to recombine

        Returns:
            list: the recombined population
        """
        p_recombination = population[:int(0.1*self.search_params['n_pop'])]
        for pop in p_recombination:
            domain = range(len(p_recombination))
            p_rec = p_recombination[self.random_state.choice(domain)][0]
            population.append([pop[0].recombine(p_rec), pop[1], pop[2], np.inf])

        return population

    def execute_optimization(self, m_max, nn_params, X, y):
        """
        Execute the optimization.

        Args:
            m_max (int): the maximum number of hidden units
            nn_params (dict): the nearest neighbors parameters
            X (np.array): all vectors
            y (np.array): all target labels

        Returns:
            list: the best constituent of the final population
        """
        n_pop = self.search_params['n_pop']
        n_iter = self.search_params['n_iter']

        smote = SMOTE(proportion=self.proportion,
                              n_neighbors=self.n_neighbors,
                              nn_params=nn_params,
                              ss_params=self.ss_params,
                              n_jobs=self.n_jobs,
                              random_state=self._random_state_init)

        X_orig, y_orig = X, y

        X, y = smote.sample(X, y)

        # setting epoch lengths
        epoch_len = int(self.search_params['n_iter']\
                        / self.search_params['n_sampling_epoch'])

        # generating initial population

        population = [self.init_pop(X=X,
                                    domain=np.arange(X.shape[1]),
                                    n_random=int(X.shape[1]/2),
                                    m_max=m_max) \
                            for _ in range(self.search_params['n_init_pop'])]
        population = [[pop, X, y, np.inf] for pop in population]
        population = sorted([[pop[0], pop[1], pop[2], pop[0].evaluate(pop[1], pop[2])]
                             for pop in population], key=lambda x: x[3])
        population = population[:n_pop]

        # executing center improval in the hidden units
        for pop in population:
            pop[0].improve_centers()

        # executing the optimization process
        for iteration in range(n_iter):
            _logger.info("%s: Iteration %d/%d, loss: %f, data size: %d",
                        self.__class__.__name__, iteration,
                        n_iter, population[0][3], len(population[0][1]))

            # evaluating non-evaluated elements
            for pop in population:
                if pop[3] == np.inf:
                    pop[3] = pop[0].evaluate(pop[1], pop[2])

            # sorting the population by the loss values
            population = sorted((pop for pop in population), key=lambda x: x[3])
            population = population[:n_pop]

            # executing mutation
            for pop in population[:int(0.1*n_pop)]:
                population.append([pop[0].mutation(), pop[1], pop[2], np.inf])

            # executing structural mutation
            for pop in population[:int(0.9*n_pop-1)]:
                population.append(
                    [pop[0].structural_mutation(), pop[1], pop[2], np.inf])

            # executing recombination
            self.recombination(population)

            # do the sampling
            if iteration % epoch_len == 0:
                X, y = smote.sample(X_orig, y_orig)
                extension = []
                for pop in population:
                    extension.append([pop[0].clone()\
                                        .update_data(X)\
                                        .improve_centers(), X, y, np.inf])
                population.extend(extension)

        # evaluate unevaluated elements of the population
        for pop in population:
            if pop[3] == np.inf:
                pop[3] = pop[0].evaluate(pop[1], pop[2])

        # sorting the population
        population = sorted((pop for pop in population),
                            key=lambda x: x[3])[:n_pop]

        return population[0]

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "Sampling is not needed")

        m_max = min(len(X), self.params['m_max'])

        if self.params['m_min'] >= m_max or len(X) < self.params['m_min'] + 1:
            return self.return_copies(X, y,
                    "Range of the number of hidden units is not suitable "\
                    f"{self.descriptor()}")

        # Standardizing the data to let the network work with comparable
        # attributes

        scaler = StandardScaler()
        X = scaler.fit_transform(X)

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        p_best = self.execute_optimization(m_max, nn_params, X, y)

        return scaler.inverse_transform(p_best[1]), p_best[2]

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'ss_params': self.ss_params,
                'hidden_range': (self.params['m_min'], self.params['m_max']),
                'i_b': self.params['i_b'],
                'o_b': self.params['o_b'],
                'n_pop': self.search_params['n_pop'],
                'n_init_pop': self.search_params['n_init_pop'],
                'n_iter': self.search_params['n_iter'],
                'n_sampling_epoch': self.search_params['n_sampling_epoch'],
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the metric tensor related functionalities.
"""

import numpy as np

import pytest

from smote_variants.base import (NearestNeighborsWithMetricTensor,
                            ClassifierImpliedDissimilarityMatrix,
                            MetricTensor,
                            MetricLearningMixin,
                            pairwise_distances_mahalanobis,
                            fix_pd_matrix,
                            reverse_matrix,
                            construct_tensor,
                            ClosestNeighborsInClasses,
                            RemoveCorrelatedColumns,
                            discrete_variable_mask,
                            estimate_mutual_information,
                            n_neighbors_func,
                            psd_mean,
                            distances_mahalanobis)

def test_distances_mahalanobis():
    """
    Test the distances_mahalanobis functions.
    """
    X = np.array([[1.0, 0.0], [0.0, 1.0]])
    assert distances_mahalanobis(X, X).shape[0] == 2

def test_pairwise_distances_mahalanobis():
    """
    Testing the pairwise distance calculation.
    """
    X = np.array([[1.0, 0.0], [0.0, 1.0]])
    Y = np.array([[2.0, 0.0]])

    dmatrix = pairwise_distances_mahalanobis(X, Y=X)
    assert dmatrix.shape == (2, 2)
    assert dmatrix[0, 0] == 0.0
    np.testing.assert_almost_equal(dmatrix[1, 0], np.sqrt(2))

    dmatrix = pairwise_distances_mahalanobis(X)
    assert dmatrix.shape == (2, 2)
    assert dmatrix[0, 0] == 0.0
    np.testing.assert_almost_equal(dmatrix[1, 0], np.sqrt(2))

    dmatrix = pairwise_distances_mahalanobis(X, Y=Y)
    assert dmatrix.shape == (2, 1)

    dmatrix_2 = pairwise_distances_mahalanobis(X, Y=Y, tensor=np.eye(2))
    np.testing.assert_array_equal(dmatrix, dmatrix_2)

def test_fix_pd_matrix():
    """
    Testing the fixing of positive definite matrices.
    """
    matrix = np.eye(2)
    matrix[1, 1]= -1
    matrix_fixed = fix_pd_matrix(matrix, eps=1e-4)

    np.testing.assert_almost_equal(matrix_fixed[1, 1], 1e-4)

def test_reverse_matrix():
    """
    Testing the reversion of matrices.
    """
    matrix = np.eye(2)
    matrix[1, 1]= -1
    matrix_reverted = reverse_matrix(matrix)

    assert matrix_reverted[0, 0] == -1

def test_construct_tensor():
    """
    Testing the tensor construction.
    """
    X = np.array([[1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    dissim_matrix = np.array([[0.0, np.sqrt(2.0), 1.0],
                                [np.sqrt(2.0), 0.0, 1.0],
                                [1.0, 1.0, 0.0]])
    tensor = construct_tensor(X, dissim_matrix)

    np.testing.assert_array_almost_equal(tensor[0], np.eye(2))

    X = np.vstack([X]*200)
    dissim_matrix = pairwise_distances_mahalanobis(X)
    tensor = construct_tensor(X, dissim_matrix)

    np.testing.assert_array_almost_equal(tensor[0], np.eye(2))

def test_classifierimplied_dissimmatrix():
    """
    Testing the classifier implied dissimilarity mixin.
    """
    cidm = ClassifierImpliedDissimilarityMatrix()
    X = np.array([[1.0, 0.0], [0.0, 1.0], [1.0, 1.0],
                    [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    y = np.array([0, 0, 1, 0, 0, 1])
    cidm.fit(X, y)

    dissim = cidm.dissimilarity_matrix(X[:3])

    np.testing.assert_almost_equal(dissim, np.array([[0.  , 0.38, 0.16],
                                            [0.38, 0.  , 0.23],
                                            [0.16, 0.23, 0.  ]]))

    assert len(cidm.get_params()) > 0

    assert len(cidm.transform_kneighbors(X)) == len(X)

def test_closest_neighbors_in_classes():
    """
    Testing the closest neighbors across classes determination.
    """
    X = np.array([[1.0, 0.0], [0.0, 1.0], [1.0, 1.0], [2.0, 2.0]])
    y = np.array([0, 0, 1, 1])

    cnic = ClosestNeighborsInClasses(n_neighbors = 2)

    _, y_final = cnic.fit_transform(X, y)

    assert np.sum(y_final == 1) == 1
    assert len(cnic.get_params()) > 0

def test_remove_correlated_columns():
    """
    Testing the removal of correlated columns.
    """
    X = np.array([[1.0, 1.0], [0.0, 0.0], [1.0, 1.0], [2.0, 2.0]])
    rcc = RemoveCorrelatedColumns()
    assert rcc.fit(X).transform(X).shape[1] == 1

def test_discrete_variable_mask():
    """
    Testing the detection of discrete variable masks.
    """
    X = np.array([[1.0, 1.0], [0.0, 0.0], [3.0, 0.0], [2.0, 1.0]])
    np.testing.assert_array_equal(discrete_variable_mask(X, threshold=2),
                                     np.array([False, True]))

def test_estimate_mutual_information():
    """
    Testing the mutual information estimation.
    """
    X = np.array([[0.0, 1.0], [1.0, 2.0], [0.0, 1.0],
                [0.0, 2.0], [1.0, 0.0],
                [1.0, 2.0], [0.0, 1.0], [1.0, 0.0]])
    y = np.array([0, 1, 0, 0, 1, 1, 0, 1])

    mutuali = estimate_mutual_information(X, y, n_repeats=2)

    assert mutuali[0] > mutuali[1]

    mutuali = estimate_mutual_information(X, y, n_repeats=2, normalize=False)

    assert mutuali[0] > mutuali[1]

def test_n_neighbors_func():
    """
    Testing the n_neighbors functionality.
    """
    X_base = np.array([[1.0, 0.0], [0.0, 1.0]])
    X_neighbors = np.array([[2.0, 0.0], [0.0, 2.0]])

    ind = n_neighbors_func(X_base, X_neighbors, n_neighbors=2)

    assert ind[0][0] == 0

    dist, ind = n_neighbors_func(X_base, X_neighbors,
                            n_neighbors=2, return_distance=True)

    assert dist[0][0] == 1.0

def test_psd_mean():
    """
    Testing the averaging of postivie definite matrices.
    """
    matrices = [np.array([[1.0, 0.0], [0.0, 1.0]]),
                np.array([[2.0, 0.0], [0.0, 1.0]])]

    etalon = np.array([[1.5, 0.0], [0.0, 1.0]])

    np.testing.assert_array_almost_equal(psd_mean(matrices),
                                            etalon)

def test_metrictensor():
    """
    Testing the metric tensor object.
    """
    metrict = MetricTensor()
    assert metrict.get_params()['metric_learning_method'] is None

@pytest.mark.parametrize('method', ['ITML', 'rf', 'LSML', 'cov',
                                    'cov_min', 'MI_weighted', 'id',
                                    'ITML_mi', 'NCA', 'gmean'])
def test_metrictensor_learning(method):
    """
    Testing the metric learning

    Args:
        method (str): metric learning algorithm
    """
    X = np.array([[0.0, 1.0], [1.5, 2.0], [0.5, 1.0],
                [0.0, 2.0], [1.0, 0.0],
                [1.0, 2.0], [0.0, 1.5], [1.5, 0.0]])
    y = np.array([0, 1, 0, 0, 1, 1, 0, 1])

    metrict = MetricTensor(metric_learning_method=method)
    assert metrict.tensor(X, y).shape == (2, 2)

def test_metriclearningmixin():
    """
    Testing the metric learning mixin.
    """
    X = np.array([[0.0, 1.0], [1.5, 2.0], [0.5, 1.0],
                [0.0, 2.0], [1.0, 0.0],
                [1.0, 2.0], [0.0, 1.5], [1.5, 0.0]])
    y = np.array([0, 1, 0, 0, 1, 1, 0, 1])

    mlm = MetricLearningMixin()

    nn_params = {'metric_learning_method': 'cov',
                    'metric': 'precomputed'}

    tensor = mlm.metric_tensor_from_nn_params(nn_params, X, y)
    assert tensor.shape == (2, 2)

    nn_params = {'metric_tensor': np.eye(2)}
    tensor = mlm.metric_tensor_from_nn_params(nn_params, X, y)
    assert tensor.shape == (2, 2)

    nn_params = {'metric': 'minkowski',
                'metric_tensor': None}
    assert mlm.metric_tensor_from_nn_params(nn_params, X, y) is None

    assert len(mlm.get_params()) == 0

def test_nn_metric_tensor():
    """
    Testing the nearest neighbors with metric tensor.
    """
    X_base = np.array([[1.0, 0.0], [0.0, 1.0]])
    X_neighbors = np.array([[2.0, 0.0], [0.0, 2.0]])

    nearestn = NearestNeighborsWithMetricTensor(n_neighbors=2)

    nearestn.fit(X_neighbors)

    ind = nearestn.kneighbors(X_base, return_distance=False)

    assert ind[0][0] == 0

    dist, ind = nearestn.kneighbors(X_base)

    assert dist[0][0] == 1.0

    ind = nearestn.radius_neighbors(X_base, radius=1.2,
                                        return_distance=False)

    assert len(ind[0]) == 1

    dist, ind = nearestn.radius_neighbors(X_base, radius=1.2)

    assert dist[0][0] == 1.0

    nearestn = NearestNeighborsWithMetricTensor(n_neighbors=2,
                                                metric='precomputed',
                                                metric_tensor=np.eye(2))

    nearestn.fit(X_neighbors)

    ind = nearestn.kneighbors(X_base, return_distance=False)

    assert ind[0][0] == 0

    dist, ind = nearestn.kneighbors(X_base)

    assert dist[0][0] == 1.0

    ind = nearestn.radius_neighbors(X_base, radius=1.2,
                                        return_distance=False)

    assert len(ind[0]) == 1

    dist, ind = nearestn.radius_neighbors(X_base, radius=1.2)

    assert dist[0][0] == 1.0

    assert len(nearestn.get_params()) > 0
